from dotmap import DotMap
import numpy as np

from ..materials.material_collection import MaterialCollection


def print_cfg(cfg: DotMap):
    """Prints the config in a more readable way.

    Args:
        cfg (DotMap): Config to print.
    """
    # Print the config three values per line
    idx = 0
    for key, value in cfg.items():
        if isinstance(value, list) or isinstance(value, np.ndarray):
            print()
            print(f"{key}: {value}")
            idx = 0
        elif isinstance(value, MaterialCollection):
            print()
        else:
            if idx % 3 == 2:
                print(f"{key:<23}: {value:<15}|")
            else:
                print(f"{key:<23}: {value:<15}|", end="")
        idx += 1
    print()



from dotmap import DotMap
import torch
from loguru import logger

from .init_trcwa import _init_trcwa


def compute_spectrum_trcwa(eps_grid, run_cfg: DotMap):
    """Evaluates TRCWA for the given epsilon values and run configuration.

    Args:
        eps_grid (torch.tensor): Epsilon values to run TRCWA with.
        run_cfg (DotMap): Run configuration.

    Raises:
        ValueError: Raises an error if epsilon contains NaN values

    Returns:
        tuple of lists: Reflection coefficients, transmission coefficients for the given epsilon values and frequencies
    """
    logger.trace("Computing spectrum")
    produced_R_spectrum = []
    produced_T_spectrum = []

    logger.debug("Testing input for NaNs")
    # Check for NaNs in eps_grid
    if torch.any(torch.isnan(eps_grid)):
        raise ValueError("Model output became NaN...", eps_grid)

    logger.debug("Checking frequencies were passed")
    assert "target_frequencies" in run_cfg, "No frequencies were passed."
    assert len(run_cfg.target_frequencies) == run_cfg.N_freq

    logger.debug("Checking passed eps_grid shape matches cfg shape")
    assert eps_grid.shape[0] == run_cfg.Nx
    assert eps_grid.shape[1] == run_cfg.Ny
    assert eps_grid.shape[2] == run_cfg.N_layers
    assert eps_grid.shape[3] == run_cfg.N_freq

    logger.debug("Iterating over passed frequencies")
    # For each frequency, evaluate TRCWA and with the grid of epsilon values
    for idx, freq in enumerate(run_cfg.target_frequencies):

        # Create TRCWA Object for this frequency
        trcwa = _init_trcwa(eps_grid[:, :, :, idx], freq, run_cfg)

        # Compute the spectrum
        reflectance, transmittance = trcwa.RT_Solve(normalize=1)

        produced_R_spectrum.append(reflectance)
        produced_T_spectrum.append(transmittance)

    return produced_R_spectrum, produced_T_spectrum



from loguru import logger
import torch


def _spectrum_loss_fn(
    produced_R_spectrum,
    produced_T_spectrum,
    target_R_spectrum,
    target_T_spectrum,
    target_frequencies,
    L=2.0,
    include_absorption=False,
):
    """Computes the loss for the difference in reflectance,
    transmittance and optionally absorption spectra.

    Args:
        produced_R_spectrum (torch.tensor): Produced reflectance spectrum.
        produced_T_spectrum (torch.tensor): Produced transmittance spectrum.
        target_R_spectrum (torch.tensor): Target reflectance spectrum.
        target_T_spectrum (torch.tensor): Target transmittance spectrum.
        target_frequencies (list of float): Target frequencies.
        L (float, optional): Exponent of the loss, e.g. L=2 for squared loss values. Defaults to 0.5.
        include_absorption (bool, optional): If absorption should be included in loss term. Defaults to False.

    Returns:
        [torch.tensor], renormalized: Loss values for reflectance, transmittance and optionally absorption and if renormalization happened.
    """
    logger.trace("Computing spectrum loss..")

    R_loss, T_loss, A_loss = 0, 0, 0
    # Tracks if renormalization happened
    renormalized = False

    # Iterate over all frequencies
    for prod_R, prod_T, target_R, target_T, freq in zip(
        produced_R_spectrum,
        produced_T_spectrum,
        target_R_spectrum,
        target_T_spectrum,
        target_frequencies,
    ):

        # If the unphysical scenario of negative absorption, i.e.
        # reflection + transmission > 1, then renormalize them to 1 and warn
        if prod_R + prod_T > 1.0:
            logger.warning(
                f"R+T>1 for freq={freq:.4f}. Renormalizing this point in the loss..."
            )
            prod_R = prod_R / (prod_R + prod_T)
            prod_T = prod_T / (prod_R + prod_T)
            renormalized = True

        # Accumulate losses
        R_loss += torch.abs(prod_R - target_R) ** L
        T_loss += torch.abs(prod_T - target_T) ** L

        if include_absorption:
            prod_A = 1 - prod_T - prod_R
            target_A = 1 - target_T - target_R
            A_loss += torch.abs(prod_A - target_A) ** L

    if include_absorption:
        return (A_loss + R_loss + T_loss) / (3 * len(produced_R_spectrum)), renormalized
    else:
        return (R_loss + T_loss) / (2 * len(produced_R_spectrum)), renormalized



from ..materials.material_collection import MaterialCollection


def test_material_collection_init():
    """Tests if the material collection can be initialized successfully."""
    target_frequencies = [9.5, 1.0, 0.1, 0.01]
    mc = MaterialCollection(target_frequencies)
    assert len(mc.material_names) > 0
    assert mc.target_frequencies == target_frequencies
    assert mc.epsilon_matrix is not None
    assert mc.epsilon_matrix.shape[0] == len(mc.material_names)
    assert mc.epsilon_matrix.shape[1] == len(target_frequencies)


if __name__ == "__main__":
    test_material_collection_init()



import pandas
import torch


def _load_material_data(name, target_frequencies=[1.0]):
    """Loads the passed wavelength, n, k data from the passed csv file for the closest frequencies and returns epsilon (permittivity).

    Args:
        name (str): Path to csv.
        target_frequencies (list, optional): Target frequencies to compute epsilon for. Defaults to [1.0].

    Returns:
        torch.tensor: Epsilon (complex permittivity) for the material
    """
    csv_data = pandas.read_csv(name, delimiter="\t")

    eps = []
    for freq in target_frequencies:
        wl = 1.0 / freq
        result_index = csv_data["Wavelength"].sub(wl).abs().idxmin()
        entry = csv_data.iloc[result_index]
        real = entry.n * entry.n - entry.k * entry.k
        imag = 2 * entry.n * entry.k
        eps.append([real + imag * 1.0j])

    eps = torch.tensor(eps)
    return eps



from ..training.run_training import run_training
from ..utils.load_default_cfg import load_default_cfg


def _setup():
    # Load default cfg as starting point
    cfg = load_default_cfg()

    # Specify your desired range of wavelengths
    cfg.physical_wavelength_range[0] = 2e-6
    cfg.physical_wavelength_range[1] = 1e-5

    # Let's investigate 4 frequency points
    cfg.N_freq = 2

    # Currently, the target spectra is set manually as a list of numbers
    cfg.target_reflectance_spectrum = [0.75, 0.25]
    cfg.target_transmittance_spectrum = [0.25, 0.25]

    cfg.Nx = 1  # Set layer size  to 1x1 (interpreted as uniform)
    cfg.Ny = 1
    cfg.N_layers = 1  # Choose number of layers
    cfg.eps_oversampling = 1

    # Allowed range of epsilon values
    cfg.real_min_eps = 0.01
    cfg.real_max_eps = 20.0
    cfg.imag_min_eps = 0.0
    cfg.imag_max_eps = 1.0

    cfg.type = "regression"
    cfg.iterations = 3  # Set number of training iterations (that is forward model evaluations) to perform

    return cfg


def test_rcwa_training():
    """Tests the training with RCWA."""
    cfg = _setup()
    cfg.solver = "TRCWA"
    run_training(cfg)
    # Should have at least learned something
    assert cfg.results.loss_log[-1] < 0.3
    cfg.pop("model", None)
    # Forget the old model
    cfg.type = "classification"
    run_training(cfg)
    # Should have at least learned something
    assert cfg.results.loss_log[-1] < 0.3


def test_fdtd_training():
    cfg = _setup()
    cfg.solver = "FDTD"
    run_training(cfg)
    # Should have at least learned something
    assert cfg.results.loss_log[-1] < 0.3
    cfg.pop("model", None)
    # Forget the old model
    cfg.type = "classification"
    run_training(cfg)
    # Should have at least learned something
    assert cfg.results.loss_log[-1] < 0.3



from dotmap import DotMap
from loguru import logger
import torch

from ..fdtd import (
    AbsorbingObject,
    set_backend,
    Grid,
    PML,
    PeriodicBoundary,
    LineSource,
    LineDetector,
    PointSource,
)
from ..utils.global_constants import EPS_0, PI, SPEED_OF_LIGHT, UNIT_MAGNITUDE
from .constants import FDTD_GRID_SCALE


def init_fdtd(cfg: DotMap, include_object, wavelength, permittivity):
    """Initialize the FDTD grid, with detectors, source, boundaries and optional objects

    Args:
        cfg (DotMap): configuration
        include_object (bool): Whether an object should be added to the grid or not
        wavelength (float): Wavelength to be used by the source
        permittivity (array): array of permittivities of the object for each layer

    Returns:
        fdtd:Grid: Grid with all the added object, ready to be run
    """
    set_backend("torch")
    if (len(cfg.PER_LAYER_THICKNESS) == 1) and (cfg.N_layers > 1):
        cfg.PER_LAYER_THICKNESS = cfg.PER_LAYER_THICKNESS * cfg.N_layers
    thicknesses = torch.tensor(cfg.PER_LAYER_THICKNESS)
    # The scaling is the number of grid points per unit magnitude. This is the maximum of the relation between the unit magnitude and 1/10th of the smallest wavelength,
    # and a constant which is defaulted to 10. If this scaling becomes too low, i.e. below 2, there might be some errors in creating the grid,
    # as there are too few grid points for certain elements to be placed correctly.
    scaling = torch.maximum(
        torch.tensor(
            UNIT_MAGNITUDE / (cfg.physical_wavelength_range[0] * FDTD_GRID_SCALE)
        ),
        torch.tensor(cfg.FDTD_min_gridpoints_per_unit_magnitude),
    )

    x_grid_size = (
        scaling
        * (
            cfg.FDTD_pml_thickness * 2
            + cfg.FDTD_free_space_distance * 2
            + torch.sum(thicknesses)
        )
    ).int()

    y_grid_size = 3
    logger.debug(
        "Initializing FDTD grid with size {} by {} grid points, with a scaling factor of {} grid points per um".format(
            x_grid_size, y_grid_size, scaling
        )
    )
    grid = Grid(
        (x_grid_size.item(), y_grid_size, 1),
        grid_spacing=UNIT_MAGNITUDE / scaling,
        permittivity=1.0,
        permeability=1.0,
    )
    grid = _add_boundaries(grid, int(cfg.FDTD_pml_thickness * scaling))

    # Determine detector positions
    detector_x = (
        cfg.FDTD_pml_thickness + cfg.FDTD_free_space_distance + torch.sum(thicknesses)
    )

    detector_y = torch.tensor(cfg.FDTD_pml_thickness + cfg.FDTD_free_space_distance)
    detector_x *= scaling
    detector_y *= scaling
    detector_x += cfg.FDTD_gridpoints_from_material_to_detector
    detector_y -= cfg.FDTD_gridpoints_from_material_to_detector

    logger.trace(
        "Placing detectors at "
        + str(detector_x.int().item())
        + ","
        + str(detector_y.int().item())
    )

    grid, t_detector, r_detector = _add_detectors(
        grid,
        detector_x.int().item(),
        detector_y.int().item(),
    )

    source_x = (cfg.FDTD_source_position[0] * scaling).int()
    source_y = (cfg.FDTD_source_position[1] * scaling).int()

    logger.trace(
        "Placing source at "
        + str(source_x.int().item())
        + ","
        + str(source_y.int().item())
    )

    grid = _add_source(
        grid,
        source_x.item(),
        source_y.item(),
        wavelength / SPEED_OF_LIGHT,
        cfg.FDTD_pulse_type,
        cfg.FDTD_source_type,
    )

    if include_object:
        x_start = torch.tensor(cfg.FDTD_pml_thickness + cfg.FDTD_free_space_distance)
        x_end = x_start + thicknesses[0]
        for i in range(permittivity.shape[2]):
            # TODO: Implement possibility for patterned grid, currently uniform layer is used
            grid = _add_object(
                grid,
                (scaling * x_start).int(),
                (scaling * x_end).int(),
                permittivity[0][0][i],
                frequency=SPEED_OF_LIGHT / wavelength,
            )

            if i < permittivity.shape[2] - 1:
                x_start += thicknesses[i]
                x_end += thicknesses[i + 1]
            else:
                break

    return grid, t_detector, r_detector


def _add_boundaries(grid, pml_thickness):
    """Add the desired boundaries to teh grid

    Args:
        grid (fdtd.Grid): The grid object which to add the boundaries
        pml_thickness (float): Thickness of the PML boundaries,

    Returns:
        fdtd.Grid: The grid with the added boundaries
    """
    # Add PML boundary to the left side of the grid
    grid[0:pml_thickness, :, :] = PML(name="pml_xlow")
    # Add PML boundary to the right side of the grid
    grid[-pml_thickness:, :, :] = PML(name="pml_xhigh")
    # Add periodic boundaries at both sides in y-direction
    grid[:, 0, :] = PeriodicBoundary(name="ybounds")
    # Add periodic boundaries on both sides in z-direction. Only applicable for 3D grids
    # grid[:, :, 0] = fdtd.PeriodicBoundary(name="zbounds")
    return grid


def _add_source(grid, source_x, source_y, period, signal_type, source_type):
    """Add a specified source to the fdtd grid

    Args:
        grid (fdtd.Grid): The grid object which to add the source
        source_x (float): The x coordinates of the placement of the source
        source_y (float): The y coordinates of the placement of the source
        period (float): The period which the source should use, given by wavelength / speed_of_light
        use_pulse_source (bool): True if pulse source is desired, Fasle to use continuous wave
        use_point_source (bool): True if a point source is desired, False if a LineSource is desired

    Returns:
        fdtd.Grid: The grid with the added source
    """
    assert signal_type in ["continuous", "hanning", "ricker"]
    if source_type == "point":
        grid[source_x, source_y, 0] = PointSource(
            period=period,
            name="pointsource",
            signal_type=signal_type,
            cycle=5,
            hanning_dt=1e-15,
        )
    elif source_type == "line":
        grid[source_x, :, 0] = LineSource(
            period=period,
            name="linesource",
            signal_type=signal_type,
            cycle=5,
            hanning_dt=1e-15,
        )
    else:
        raise ValueError(f'FDTD_source_type must be either "line" or "point"')
    return grid


def _add_detectors(grid, transmission_detector_x, reflection_detector_x):
    """Add a transmission detector and reflection detector to the fdtd grid

    Args:
        grid (fdtd.Grid): The grid object which to add the detectors
        transmission_detector_x (float): x coordinates of the transmission detector
        reflection_detector_x (float): x coordinates of teh reflection detector

    Returns:
        fdtd.Grid: The grid with the added detectors
    """

    transmission_detector = LineDetector(name="t_detector")
    reflection_detector = LineDetector(name="r_detector")
    grid[transmission_detector_x, :, 0] = transmission_detector
    grid[reflection_detector_x, :, 0] = reflection_detector
    return grid, transmission_detector, reflection_detector


def _add_object(grid, object_start_x, object_end_x, permittivity, frequency):
    """Add a object to the fdtd grid, with a specified permittivity. The object covers the entire grid in the y-direction.

    Args:
        grid (fdtd:Grid): The grid object which to add the object
        object_start_x (float): Coordinate where the object should start in the x-direction
        object_end_x (float): Coordinate where the object should end in the x-direction
        permittivity (float): The dielectric permittivity of the object

    Returns:
        fdtd.Grid: The grid with the added object
    """
    # Not sure whether the conductivity should be relative or absolute, i.e. if it should be multiplied with EPS_0. Multiplied with 2pi to get w(angular frequency)?
    # Since the permittivity is set to 1 for the free space grid, I'll leave it at an relative value for now. Also, permittivity for object is relative.
    grid[object_start_x:object_end_x, :, :] = AbsorbingObject(
        permittivity=permittivity.real,
        conductivity=permittivity.imag * frequency * 2 * PI * EPS_0,
    )
    return grid



from dotmap import DotMap
from loguru import logger

from nidn.fdtd_integration.constants import FDTD_GRID_SCALE
from ..utils.global_constants import UNIT_MAGNITUDE


def _validate_config(cfg: DotMap):
    """This function validates that all required entries are in the config.

    Args:
        cfg (DotMap): Run config you intend to use.
    """
    _check_for_keys(cfg)
    _check_entry_types(cfg)
    _check_value_ranges(cfg)
    if cfg.solver == "FDTD":
        _check_grid_scale(cfg)
    logger.debug("Config validated successfully.")


def _check_grid_scale(cfg):
    """Checks whether FDTD grid scale allows specified sizes.

    Args:
        cfg (DotMap): Configurations for the simulation
    """
    scale_in_wl = cfg.physical_wavelength_range[0] * FDTD_GRID_SCALE
    scaling = max(
        UNIT_MAGNITUDE / scale_in_wl,
        cfg.FDTD_min_gridpoints_per_unit_magnitude,
    )
    for i, thickness in enumerate(cfg.PER_LAYER_THICKNESS):
        x_start = cfg.FDTD_pml_thickness + cfg.FDTD_free_space_distance
        x_end = x_start + thickness
        size = int(scaling * x_end) - int(scaling * x_start)
        if thickness != size / scaling:
            downscaled = size / scaling
            logger.warning(
                f"Due to the grid resolution, the thickness of layer {i + 1} is set to {downscaled:.3f} µm"
                + f"instead of the specified {thickness} µm"
            )


def _check_for_keys(cfg: DotMap):
    """Checks that all required keys are present in the config"""
    # fmt: off
    required_keys = ["name","seed","eps_oversampling","real_min_eps","real_max_eps","imag_min_eps","imag_max_eps",
                     "Nx","Ny","N_layers","physical_wavelength_range","N_freq","model_type","encoding_dim","siren_omega",
                     "n_neurons","hidden_layers","learning_rate","iterations","L","absorption_loss","type","use_regularization_loss",
                     "reg_loss_weight","add_noise","noise_scale","solver","PER_LAYER_THICKNESS",
                     "TRCWA_L_grid","TRCWA_NG","TRCWA_TOP_LAYER_EPS","TRCWA_BOTTOM_LAYER_EPS",
                     "FDTD_min_gridpoints_per_unit_magnitude","FDTD_source_type","FDTD_pulse_type","FDTD_pml_thickness",
                     "FDTD_source_position","FDTD_free_space_distance","FDTD_niter","FDTD_gridpoints_from_material_to_detector",
                     "target_reflectance_spectrum","target_transmittance_spectrum","freq_distribution","use_gpu","avoid_zero_eps"]

    # Some keys that may be set in the cfg during runtime
    optional_keys = ["target_frequencies","FDTD_grid","model","out_features","results","best_model_state_dict","material_collection",
                     "N_materials"]
    # fmt: on
    for key in required_keys:
        if key not in cfg:
            raise KeyError(f"CFG missing required key: {key}")

    for key in cfg.keys():
        if key not in required_keys and key not in optional_keys:
            raise KeyError(f"CFG Key {key} is not a valid key")


def _check_entry_types(cfg: DotMap):
    """Check that all entries in the config are of the correct type"""
    # fmt: off
    integer_keys = ["Nx","Ny","N_layers","N_freq","encoding_dim","n_neurons","eps_oversampling","seed",
                    "TRCWA_NG","FDTD_niter","FDTD_gridpoints_from_material_to_detector","FDTD_min_gridpoints_per_unit_magnitude"]
    float_keys = ["L","real_min_eps","real_max_eps","imag_min_eps","imag_max_eps","siren_omega","noise_scale","reg_loss_weight",
                  "TRCWA_TOP_LAYER_EPS","TRCWA_BOTTOM_LAYER_EPS","FDTD_free_space_distance","FDTD_pml_thickness",]
    boolean_keys = ["use_regularization_loss","add_noise","use_gpu","avoid_zero_eps",]
    string_keys = ["model_type","type","name","freq_distribution","solver","FDTD_source_type","FDTD_pulse_type",]
    list_keys = ["PER_LAYER_THICKNESS","TRCWA_L_grid","FDTD_source_position","target_reflectance_spectrum","target_transmittance_spectrum",
                 "physical_wavelength_range",]
    # fmt: on

    for key in integer_keys:
        if not isinstance(cfg[key], int):
            raise TypeError(f"{key} must be an integer")

    for key in float_keys:
        if not isinstance(cfg[key], float):
            raise TypeError(f"{key} must be a float")

    for key in boolean_keys:
        if not isinstance(cfg[key], bool):
            raise TypeError(f"{key} must be a boolean")

    for key in string_keys:
        if not isinstance(cfg[key], str):
            raise TypeError(f"{key} must be a string")

    for key in list_keys:
        if not isinstance(cfg[key], list):
            raise TypeError(f"{key} must be a list")


def _check_value_ranges(cfg: DotMap):
    """Check that all values in the config are within the correct range.
    This throws runtime errors as ValueErrors are caught in training to avoid NaNs crashing the training."""
    if not (cfg.real_min_eps <= cfg.real_max_eps):
        raise RuntimeError(f"real_min_eps must be less than or equal to real_max_eps")

    if not (cfg.imag_min_eps <= cfg.imag_max_eps):
        raise RuntimeError(f"imag_min_eps must be less than or equal to imag_max_eps")

    if not (cfg.physical_wavelength_range[0] < cfg.physical_wavelength_range[1]):
        raise RuntimeError(
            f"physical_wavelength_range must be ordered from low to high"
        )

    scaling = max(
        UNIT_MAGNITUDE / (cfg.physical_wavelength_range[0] * FDTD_GRID_SCALE),
        cfg.FDTD_min_gridpoints_per_unit_magnitude,
    )
    if (
        int(scaling * cfg.FDTD_free_space_distance)
        < cfg.FDTD_gridpoints_from_material_to_detector
    ):
        raise RuntimeError(
            "Reflection detector must be placed in the free space before an eventual object, and after the pml layer. Decrease FDTD_gridpoints_from_material_to_detector to a value smaller than {value} fix this.".format(
                value=int(scaling * cfg.FDTD_free_space_distance)
            )
        )

    # fmt: off
    positive_value_keys = ["L","n_neurons","Nx","Ny","N_layers","N_freq","learning_rate","iterations","eps_oversampling","noise_scale",
                           "TRCWA_NG","reg_loss_weight","FDTD_niter","FDTD_free_space_distance","FDTD_pml_thickness",
                           "FDTD_gridpoints_from_material_to_detector","FDTD_min_gridpoints_per_unit_magnitude",]
    # fmt: on

    for key in positive_value_keys:
        if not (cfg[key] > 0):
            raise RuntimeError(f"{key} must be a positive integer")

    if not cfg.Nx % cfg.eps_oversampling == 0:
        raise RuntimeError(f"Nx must be a multiple of eps_oversampling")
    if not cfg.Ny % cfg.eps_oversampling == 0:
        raise RuntimeError(f"Ny must be a multiple of eps_oversampling")

    if cfg.type != "classification" and cfg.type != "regression":
        raise RuntimeError(f"type must be either 'classification' or 'regression'")

    if not cfg.TRCWA_L_grid[1][0] < cfg.TRCWA_L_grid[1][1]:
        raise RuntimeError(f"TRCWA_L_grid dim1 must be ordered from low to high")
    if not cfg.TRCWA_L_grid[0][0] > cfg.TRCWA_L_grid[0][1]:
        raise RuntimeError(f"TRCWA_L_grid dim0 must be ordered from high to low")

    all_positive_list_keys = [
        "TRCWA_L_grid",
        "PER_LAYER_THICKNESS",
        "FDTD_grid",
        "FDTD_source_position",
    ]
    all_positive_or_zero_list_keys = [
        "target_transmittance_spectrum",
        "target_reflectance_spectrum",
    ]

    for key in all_positive_list_keys:
        if not (all(cfg[key]) > 0.0):
            raise RuntimeError(f"All elements in {key} must be a positive integer")
    for key in all_positive_or_zero_list_keys:
        if not (all(cfg[key]) >= 0.0):
            raise RuntimeError(
                f"All elements in {key} must be a positive integer or zero"
            )

    # Disabled this check temporarily as it is only relevant in training. For forward model this doesn't matter.
    # if not len(cfg.target_transmittance_spectrum) == cfg.N_freq:
    #     raise RuntimeError(
    #         f"target_transmittance_spectrum must have length N_freq. Got {len(cfg.target_transmittance_spectrum)}, expected {cfg.N_freq}"
    #     )

    # if not len(cfg.target_reflectance_spectrum) == cfg.N_freq:
    #     raise RuntimeError(
    #         f"target_reflectance_spectrum must have length N_freq. Got {len(cfg.target_reflectance_spectrum)}, expected {cfg.N_freq}"
    #     )

    if not (
        len(cfg.PER_LAYER_THICKNESS) == cfg.N_layers
        or len(cfg.PER_LAYER_THICKNESS) == 1
    ):
        raise RuntimeError(f"PER_LAYER_THICKNESS must have length 1 or N_layers")

    if not (cfg.freq_distribution == "linear" or cfg.freq_distribution == "log"):
        raise RuntimeError(f"freq_distribution must be either 'linear' or 'log'")

    if not (len(cfg.FDTD_source_position) == 2 or len(cfg.FDTD_source_position) == 3):
        raise RuntimeError(
            f"The FDTD source needs either 2- or 3-dimensional coordinates"
        )
    if not cfg.FDTD_source_type in ["point", "line"]:
        raise RuntimeError(f'The FDTD_source_type must either be "line" or "point"')

    if not cfg.FDTD_pulse_type in ["pulse", "continuous"]:
        raise RuntimeError(
            f'The FDTD_pulse_type must either be "pulse" or "continuous"'
        )



import torch
import numpy as np
from matplotlib import pyplot as plt
from matplotlib.ticker import AutoLocator
from ..utils.convert_units import freq_to_wl
from ..trcwa.compute_target_frequencies import compute_target_frequencies
from ..utils.global_constants import NIDN_PLOT_COLOR_1, NIDN_FONTSIZE


def _add_plot(
    fig,
    target_frequencies,
    spectrum,
    ylimits,
    nr,
    type_name,
    logscale=False,
    markers=True,
):
    fontsize = NIDN_FONTSIZE
    markersize = 8 if markers else 0

    ax = fig.add_subplot(nr)
    ax.plot(
        freq_to_wl(target_frequencies),
        spectrum,
        marker="o",
        c=NIDN_PLOT_COLOR_1,
        lw=2,
        markersize=markersize,
    )
    ax.legend(
        [f"Produced {type_name}", f"Target {type_name}"],
        fontsize=fontsize,
    )
    ax.set_xlabel("Wavelength [µm]", fontsize=fontsize + 2)
    ax.set_ylabel(f"{type_name}", fontsize=fontsize + 2)
    if logscale:
        ax.set_xscale("log")
    ax.xaxis.set_major_formatter(plt.FuncFormatter(lambda x, pos: "{:.1f}".format(x)))
    # ax.xaxis.set_minor_formatter(plt.FuncFormatter(lambda x, pos: "{:.1f}".format(x)))
    ax.tick_params(axis="both", which="major", labelsize=fontsize)

    ax.xaxis.set_major_locator(AutoLocator())
    plt.minorticks_off()
    # ax.axhspan(-6, 0, facecolor="gray", alpha=0.3)
    # ax.axhspan(1, 5, facecolor="gray", alpha=0.3)
    ax.set_ylim(ylimits)
    return fig


def plot_spectrum(
    run_cfg, R_spectrum, T_spectrum, markers=True, save_path=None, filename=None
):
    """Plots the produced RTA spectra. Optionally saves it.

    Args:
        run_cfg (dict): The run configuration.
        R_spectrum (torch.tensor): The reflection spectrum.
        T_spectrum (torch.tensor): The transmission spectrum.
        markers (bool): Whether to plot markers for the target and produced spectra.
        save_path (str, optional): Folder to save the plot in. Defaults to None, then the plot will not be saved.
        filename (str, optional): Filename to save the plot in. Defaults to None, then the plot will be saved with the name "spectrum.png".
    """

    if not "target_frequencies" in run_cfg.keys():
        run_cfg.target_frequencies = compute_target_frequencies(
            run_cfg.physical_wavelength_range[0],
            run_cfg.physical_wavelength_range[1],
            run_cfg.N_freq,
            run_cfg.freq_distribution,
        )
    target_frequencies = run_cfg.target_frequencies

    # Convert the spectra to numpy arrays for matplotlib
    R_spectrum = torch.tensor(R_spectrum).detach().cpu().numpy()
    T_spectrum = torch.tensor(T_spectrum).detach().cpu().numpy()

    A_spectrum = np.ones_like(np.asarray(R_spectrum)) - (
        np.asarray(T_spectrum) + np.asarray(R_spectrum)
    )

    # To align all plots
    ylimits = [0.0, 1.0]
    # if (
    #     (max(A_spectrum) > 1 or min(A_spectrum) < 0)
    #     or (max(T_spectrum) > 1 or min(T_spectrum) < 0)
    #     or (max(R_spectrum) > 1 or min(R_spectrum) < 0)
    # ):
    #     ylimits = [
    #         min(min(A_spectrum), min(T_spectrum), min(R_spectrum)) + ylimits[0],
    #         max(max(A_spectrum), max(T_spectrum), max(R_spectrum)) + 0.1,
    #     ]

    fig = plt.figure(figsize=(15, 5), dpi=300)
    fig.patch.set_facecolor("white")

    fig = _add_plot(
        fig,
        target_frequencies,
        R_spectrum,
        ylimits,
        121,
        "Reflectance",
        markers=markers,
    )
    fig = _add_plot(
        fig,
        target_frequencies,
        T_spectrum,
        ylimits,
        122,
        "Transmittance",
        markers=markers,
    )
    # fig = _add_plot(
    #     fig,
    #     target_frequencies,
    #     A_spectrum,
    #     ylimits,
    #     133,
    #     "Absorptance",
    # )

    fig.autofmt_xdate()
    plt.tight_layout()

    if save_path is not None:
        if filename is None:
            plt.savefig(save_path + "/spectrum.png", dpi=150)
        else:
            plt.savefig(save_path + "/" + filename + ".png", dpi=300)



from torch import nn
import torch


class NERFLayer(nn.Module):
    """Layer as used in NERF."""

    def __init__(self, in_features, out_features, bias=True, activation=nn.ReLU()):
        super().__init__()
        self.linear = nn.Linear(in_features, out_features, bias=bias)
        self.init_weights()
        self.activation = activation

    def init_weights(self):
        nn.init.xavier_uniform_(self.linear.weight)
        nn.init.uniform_(self.linear.bias.data, -0.0, 0.0)

    def forward(self, input):
        return self.activation(self.linear(input))


class NERF(nn.Module):
    """NERF architecture as described by Mildenhall et al. (2020)"""

    def __init__(
        self,
        in_features,
        out_features,
        skip=[],
        n_neurons=100,
        activation=nn.Sigmoid(),
        hidden_layers=8,
    ):
        super().__init__()
        self.in_features = in_features
        self.skip = skip

        self.net = nn.ModuleList()
        self.net.append(NERFLayer(in_features, n_neurons))

        for i in range(hidden_layers):
            if i in self.skip:
                self.net.append(NERFLayer(n_neurons + in_features, n_neurons))
            else:
                self.net.append(NERFLayer(n_neurons, n_neurons))

        self.net.append(NERFLayer(n_neurons, out_features, activation=activation))

    def forward(self, x):
        # save for skip connection
        identity = x

        # compute first layer
        out = self.net[0].forward(x)

        # compute all other layers and apply skip where requested
        for layer_idx in range(1, len(self.net)):
            out = self.net[layer_idx].forward(out)
            if layer_idx in self.skip:
                out = torch.cat([out, identity], dim=1)

        return out



import numpy as np

from ..utils.convert_units import phys_wl_to_wl, freq_to_wl


def compute_target_frequencies(
    min_physical_wl: float, max_physical_wl: float, N_freq: int, distribution: str
):
    """Computes the target frequencies for a given set of physical frequencies.

    Args:
        min_physical_wl (float): Minimum physical frequency in the target spectrum.
        max_physical_wl (float): Maximum physical frequency in the target spectrum.
        N_freq (int): The number of target frequencies.
        distribution (str): The distribution of the target frequencies. Either linear or log.

    Returns:
        np.array: Corresponding target frequencies
    """
    max_physical_freq = freq_to_wl(phys_wl_to_wl(min_physical_wl))
    min_physical_freq = freq_to_wl(phys_wl_to_wl(max_physical_wl))

    logval_min = np.log10(min_physical_freq)
    logval_max = np.log10(max_physical_freq)

    if distribution == "linear":
        target_frequencies = np.linspace(
            min_physical_freq, max_physical_freq, num=N_freq
        )
    elif distribution == "log":
        target_frequencies = np.logspace(logval_min, logval_max, num=N_freq, base=10)
    else:
        raise ValueError(
            "Unknown distribution type: {}. Needs to be 'log' or 'linear'.".format(
                distribution
            )
        )

    target_frequencies = np.sort(target_frequencies)
    return target_frequencies



import torch


def Epsilon_fft(dN, eps_grid, G):
    """dN = 1/Nx/Ny
    For now, assume epsilon is diagonal; if epsilon has xz,yz component, just simply add them to off-diagonal eps2

    eps_grid is  (1) for isotropic, a numpy 2d array in the format of (Nx,Ny),
                 (2) for anisotropic, a list of numpy 2d array [(Nx,Ny),(Nx,Ny),(Nx,Ny)]
    """

    if len(eps_grid) == 3 and eps_grid[0].ndim == 2:
        epsx_fft = get_conv(dN, eps_grid[0], G)
        epsy_fft = get_conv(dN, eps_grid[1], G)
        epsz_fft = get_conv(dN, eps_grid[2], G)
        epsinv = torch.inverse(epsz_fft)

        tmp1 = torch.vstack((epsx_fft, torch.zeros_like(epsx_fft)))
        tmp2 = torch.vstack((torch.zeros_like(epsx_fft), epsy_fft))
        eps2 = torch.hstack((tmp1, tmp2))

    elif eps_grid[0].ndim == 1:
        eps_fft = get_conv(dN, eps_grid, G)
        epsinv = torch.inverse(eps_fft)

        tmp1 = torch.vstack((eps_fft, torch.zeros_like(eps_fft)))
        tmp2 = torch.vstack((torch.zeros_like(eps_fft), eps_fft))
        eps2 = torch.hstack((tmp1, tmp2))
    else:
        raise ValueError("Wrong eps_grid type")

    # Set very small components to zero for torch
    for i in range(eps2.shape[0]):
        for j in range(eps2.shape[1]):
            if eps2[i][j].real < 1e-14 and eps2[i][j].imag < 1e-14:
                eps2[i][j] = 0 + 0j

    return epsinv, eps2


def get_conv(dN, s_in, G):
    """Attain convolution matrix
    dN = 1/Nx/Ny
    s_in: np.array of length Nx*Ny
    G: shape (nG,2), 2 for Lk1,Lk2
    s_out: 1/N sum a_m exp(-2pi i mk/n), shape (nGx*nGy)
    """
    nG, _ = G.shape
    sfft = torch.fft.fft2(s_in) * dN

    ix = range(nG)
    ii, jj = torch.meshgrid(torch.tensor(ix), torch.tensor(ix))
    s_out = sfft[G[ii, 0] - G[jj, 0], G[ii, 1] - G[jj, 1]]
    return s_out


def get_fft(dN, s_in, G):
    """
    FFT to get Fourier components

    s_in: np.2d array of size (Nx,Ny)
    G: shape (nG,2), 2 for Gx,Gy
    s_out: 1/N sum a_m exp(-2pi i mk/n), shape (nGx*nGy)
    """

    sfft = torch.fft.fft2(s_in) * dN
    return sfft[G[:, 0], G[:, 1]]


def get_ifft(Nx, Ny, s_in, G):
    """
    Reconstruct real-space fields
    """
    dN = 1.0 / Nx / Ny
    nG, _ = G.shape

    s0 = torch.zeros((Nx, Ny), dtype=complex)
    for i in range(nG):
        x = G[i, 0]
        y = G[i, 1]

        stmp = torch.zeros((Nx, Ny), dtype=complex)
        stmp[x, y] = 1.0
        s0 = s0 + s_in[i] * stmp

    s_out = torch.ifft2(s0) / dN
    return s_out



from dotmap import DotMap
from .compute_target_frequencies import compute_target_frequencies
from ..utils.convert_units import freq_to_wl, wl_to_phys_wl


def get_frequency_points(run_cfg: DotMap):
    """Computes both the physical and normalized frequency points which will be used.
    This can be helpful to determine the points at which you need to define the target
    spectrum.

    Args:
        run_cfg (DotMap): Config for the run.

    Returns:
        physical_target_frequencies (list): List of physical frequencies.
        target_frequencies (list): List of target frequencies.
    """

    # Determine target frequencies
    run_cfg.target_frequencies = compute_target_frequencies(
        run_cfg.physical_wavelength_range[0],
        run_cfg.physical_wavelength_range[1],
        run_cfg.N_freq,
        run_cfg.freq_distribution,
    )

    physical_target_frequencies = wl_to_phys_wl(freq_to_wl(run_cfg.target_frequencies))

    return physical_target_frequencies, run_cfg.target_frequencies



import pandas
import torch
import os, sys
from glob import glob

from loguru import logger


class MaterialCollection:
    """Database of materials included with NIDN"""

    epsilon_matrix = None
    material_names = []
    N_materials = 0

    def __init__(self, target_frequencies, to_skip=None):
        """Initalizes the MaterialCollection. Loads all materials from the data folder.

        Args:
            target_frequencies (list): Frequencies we are targeting. Closest ones in the data will be used.
            to_skip (list): List of materials to skip.
        """
        logger.trace("Initializing material collection")
        self.epsilon_matrix = None
        self.material_names = []

        if len(target_frequencies) == 0:
            raise ValueError("No target frequencies specified.")

        self.target_frequencies = target_frequencies
        self.materials_folder = os.path.dirname(__file__) + "/data/"
        self._load_materials_folder(to_skip)

    def __getitem__(self, key):
        """Given the name of a material will return the corresponding epsilon tensor.

        Args:
            key (str): Name of the material.

        Returns:
            torch.tensor: The corresponding epsilon tensor.
        """
        if key in self.material_names:
            return self.epsilon_matrix[self.material_names.index(key)]
        else:
            raise KeyError(
                f"Material '{key}' not found in the material collection. Available are: {self.material_names}"
            )

    def _load_materials_folder(self, to_skip=None):
        """Loads all csv files from folder "data"
        and sets up the EPS_MATRIX and MATERIAL_NAMES

        Args:
            to_skip (list): List of materials to skip.
        """
        logger.trace(f"Loading materials from folder: {self.materials_folder}")

        # Load materials data
        files = glob(self.materials_folder + "/*.csv")
        logger.debug("Found the following files")
        logger.debug(files)

        eps_list = []
        for file in files:
            file = file.replace("\\", "/")
            name = file.split("/")[-1].split(".csv")[0]

            # Skip materials we don't want
            if to_skip is not None and name in to_skip:
                logger.debug(f"Skipping material {name}")
                continue

            # Load material epsilon (permittivity)
            eps_list.append(self._load_material_data(file))

            # Remember file name as material name
            self.material_names.append(name)

        # Create a single tensor of all materials
        logger.trace("Creating material tensor")
        self.epsilon_matrix = torch.stack(eps_list).squeeze()  # .cuda()
        self.N_materials = len(self.material_names)

    def _load_material_data(self, name):
        """Loads data (wavelength, n, and k) from the passed csv file for the closest frequencies and returns epsilon (permittivity).

        Args:
            name (str): Path to csv.

        Returns:
            torch.tensor: Epsilon for the material (permittivity)
        """
        logger.debug(f"Loading material {name}")
        csv_data = pandas.read_csv(name, delimiter="\t")

        eps = []
        # Compute epsilon from n and k
        # see e.g. https://www.tf.uni-kiel.de/matwis/amat/elmat_en/kap_3/backbone/r3_7_2.html
        logger.trace("Computing epsilon from n and k for each target frequency")
        for freq in self.target_frequencies:
            wl = 1.0 / freq
            result_index = csv_data["Wavelength"].sub(wl).abs().idxmin()
            entry = csv_data.iloc[result_index]
            real = entry.n * entry.n - entry.k * entry.k
            imag = 2 * entry.n * entry.k
            eps.append([real + imag * 1.0j])

        eps = torch.tensor(eps)
        logger.trace(f"Epsilon for material {name} is: {eps}")
        return eps



import os
import toml
from dotmap import DotMap


def load_default_cfg():
    """Loads the default toml config file from the cfg folder."""
    path = os.path.join(
        os.path.dirname(__file__) + "/resources/", "default_config.toml"
    )
    with open(path) as cfg:
        return DotMap(toml.load(cfg))



import torch


# Q_ABS is a parameter for relaxation to better approach global optimal, at Q_ABS= inf, it will describe the real physics.
# It can also be used to resolve singular matrix issues by setting a large but finite Q_ABS, e.g. Q_ABS= 1e5.
# See page 2 of arxiv.org/pdf/2005.04840v1.pdf
# or doi.org/10.1364/OE.21.030812 for a more thorough explanation
TRCWA_Q_ABS = 1e5

# Planewave excitation
TRCWA_PLANEWAVE = {"p_amp": 0, "s_amp": 1, "p_phase": 0, "s_phase": 0}

# Angles
TRCWA_THETA = 0.0
TRCWA_PHI = 0.0



import torch
import numpy as np


def fix_random_seeds(seed: int = 42):
    """This function sets the random seeds in torch and numpy to enable reproducible behavior.

    Args:
        seed (int): Seed to use.
    """
    torch.manual_seed(seed)
    np.random.seed(seed)



import torch
from loguru import logger
from ..utils.load_default_cfg import load_default_cfg
from ..utils.compute_spectrum import compute_spectrum

_TRCWA_TEST_TOLERANCE = 1e-7
torch.set_printoptions(precision=12)
torch.set_default_tensor_type(torch.DoubleTensor)


def test_single_layer():
    """Tests a single patterned layer with a single frequency. TRCWA vs. GRCWA"""
    logger.info("Running single layer test...")
    run_cfg = load_default_cfg()
    run_cfg.Nx = 5
    run_cfg.Ny = 5
    run_cfg.N_layers = 1
    run_cfg.target_frequencies = [0.5]
    run_cfg.N_freq = 1
    run_cfg.TRCWA_L_grid = [[1.0, 0.0], [0.0, 1.0]]
    run_cfg.TRCWA_NG = 11
    run_cfg.TRCWA_TOP_LAYER_EPS = 1.0  # epsilon for top layer, 1.0 for vacuum
    run_cfg.TRCWA_BOTTOM_LAYER_EPS = 1.0  # epsilon for top layer, 1.0 for vacuum
    run_cfg.PER_LAYER_THICKNESS = [1.0]
    run_cfg.solver = "TRCWA"

    # Get eps shape
    shape = [
        run_cfg.Nx,
        run_cfg.Ny,
        run_cfg.N_layers,
        run_cfg.N_freq,
    ]

    # Get a grid of 1 + 1j
    eps_grid = (
        torch.ones(
            shape,
            dtype=torch.cfloat,
        )
        * (1.0 + 1j)
    )

    logger.debug("Computing spectrum...")

    R, T = compute_spectrum(eps_grid, run_cfg)

    # Computed with GRCWA
    R_error = torch.abs(R[0] - 0.04256584753869945)
    T_error = torch.abs(T[0] - 0.06064642157274551)

    logger.info(f"R_error = {R_error}")
    logger.info(f"T_error = {T_error}")

    assert R_error < _TRCWA_TEST_TOLERANCE
    assert T_error < _TRCWA_TEST_TOLERANCE


def test_uniform_layer():
    """Tests a single uniform layer at two frequencies. TRCWA vs. GRCWA"""
    logger.info("Running uniform layer test...")
    run_cfg = load_default_cfg()
    run_cfg.Nx = 1
    run_cfg.Ny = 1
    run_cfg.N_layers = 1
    run_cfg.target_frequencies = [0.05, 0.1]
    run_cfg.N_freq = 2
    run_cfg.TRCWA_L_grid = [[1.0, 0.0], [0.0, 1.0]]
    run_cfg.TRCWA_NG = 11
    run_cfg.TRCWA_TOP_LAYER_EPS = 1.0  # epsilon for top layer, 1.0 for vacuum
    run_cfg.TRCWA_BOTTOM_LAYER_EPS = 1.0  # epsilon for top layer, 1.0 for vacuum
    run_cfg.PER_LAYER_THICKNESS = [1.0]
    run_cfg.solver = "TRCWA"

    # Get eps shape
    shape = [
        run_cfg.Nx,
        run_cfg.Ny,
        run_cfg.N_layers,
        run_cfg.N_freq,
    ]

    # Get a grid of 3 + 1.5j
    eps_grid = torch.ones(
        shape,
        dtype=torch.cfloat,
    )

    eps_grid = eps_grid * (3.0 + 1.5j)

    logger.debug("Computing spectrum...")

    R, T = compute_spectrum(eps_grid, run_cfg)

    # Computed with GRCWA
    R_error, T_error = [], []
    R_error.append(torch.abs(R[0] - 0.08805902003703245))
    R_error.append(torch.abs(R[1] - 0.18524218068012266))
    T_error.append(torch.abs(T[0] - 0.6307193983893462))
    T_error.append(torch.abs(T[1] - 0.4510135948077737))

    logger.info(f"R_error = {R_error}")
    logger.info(f"T_error = {T_error}")

    assert all(torch.tensor(R_error) < _TRCWA_TEST_TOLERANCE)
    assert all(torch.tensor(T_error) < _TRCWA_TEST_TOLERANCE)


def test_three_layer():
    """Tests a three stacked patterned layer with a single frequency. TRCWA vs. GRCWA"""
    logger.info("Running three layer test...")
    run_cfg = load_default_cfg()
    run_cfg.Nx = 9
    run_cfg.Ny = 9
    run_cfg.N_layers = 3
    run_cfg.target_frequencies = [3.0]
    run_cfg.N_freq = 1
    run_cfg.TRCWA_L_grid = [[1.0, 0.0], [0.0, 1.0]]
    run_cfg.TRCWA_NG = 11
    run_cfg.TRCWA_TOP_LAYER_EPS = 1.0  # epsilon for top layer, 1.0 for vacuum
    run_cfg.TRCWA_BOTTOM_LAYER_EPS = 1.0  # epsilon for top layer, 1.0 for vacuum
    run_cfg.PER_LAYER_THICKNESS = [1.0]
    run_cfg.solver = "TRCWA"

    # Get eps shape
    shape = [
        run_cfg.Nx,
        run_cfg.Ny,
        run_cfg.N_layers,
        run_cfg.N_freq,
    ]

    # Get a somewhat interesting grid
    eps_grid = (
        torch.ones(
            shape,
            dtype=torch.cfloat,
        )
        * (-4.2 + 0.42j)
    )

    eps_grid[0:3, 0:3, 0, :] = eps_grid[0:3, 0:3, 0, :] * 1.0
    eps_grid[0:3, 0:3, 1, :] = eps_grid[0:3, 0:3, 1, :] * 2.0
    eps_grid[0:3, 0:3, 2, :] = eps_grid[0:3, 0:3, 2, :] * 3.0

    logger.debug("Computing spectrum...")

    R, T = compute_spectrum(eps_grid, run_cfg)

    # Computed with GRCWA
    R_error = torch.abs(R[0] - 0.9243750060320585)
    T_error = torch.abs(T[0] - 0.0)

    logger.info(f"R_error = {R_error}")
    logger.info(f"T_error = {T_error}")

    assert R_error < _TRCWA_TEST_TOLERANCE
    assert T_error < _TRCWA_TEST_TOLERANCE


if __name__ == "__main__":
    test_single_layer()
    test_uniform_layer()
    test_three_layer()



from math import sqrt

from nidn.utils.global_constants import EPS_0, MU_0

"""
Optional reduced unit conversion functions for user use.
Mainly to make it explicit where conversions happen.

Unit system described by the thesis
"Novel architectures for brain-inspired photonic computers",
https://www.photonics.intec.ugent.be/download/phd_259.pdf

Chapters 4.1.2 and 4.1.6.

> "In SI units, the relative magnitude difference between the fields is related
by the electromagnetic impedance of free space, which for the current choice
of simulation units equals 1 per definition".

FIXME: DC: find and add md notes on scaling

"""

# Might perhaps be worth putting a note in the readme about default units / suggested unit systems?
# (done, see next PR)
# Also, if /fdtd/ gets the ability to dump to VTK, one would hate to
# have to scale everything for physical results in paraview. Will put a scaling option there.
# Is it reasonable to have a global flag for scaling? probably not
#
# don't know how severe noise is with fp32, whether it's worth the extra hassle for most people


def simE_to_worldE(input):
    return input / sqrt(EPS_0)


def worldE_to_simE(input):
    return sqrt(EPS_0) * input


def simH_to_worldH(input):
    return input / sqrt(MU_0)


def worldH_to_simH(input):
    return sqrt(MU_0) * input



import torch
import numpy as np
from matplotlib import pyplot as plt

from ..materials.material_collection import MaterialCollection
from ..materials.find_closest_material import _find_closest_material
from ..training.model.model_to_eps_grid import model_to_eps_grid
from ..utils.global_constants import NIDN_FONTSIZE


def plot_material_grid(
    run_cfg, save_path=None, eps=None, plot_error=False, to_skip=None
):
    """Plots the materials closest to the used ones for each grid point. Optionally saves it.

    Args:
        run_cfg (dict): The run configuration.
        save_path (str, optional): Folder to save the plot in. Defaults to None, then the plot will not be saved.
        eps (torch.tensor, optional): The epsilon tensor. Defaults to None, then the epsilon tensor will be computed from the run_cfg.
        plot_error (bool, optional): If True, the error will be plotted. Defaults to False.
        to_skip (int, optional): Material to skip. Defaults to None.
    """
    Nx, Ny, N_layers = run_cfg.Nx, run_cfg.Ny, run_cfg.N_layers

    if eps is None:
        # Create epsilon grid from the model
        eps, _ = model_to_eps_grid(run_cfg.model, run_cfg)

    # Setup grid
    x = torch.linspace(-1, 1, Nx)
    y = torch.linspace(-1, 1, Ny)
    z = torch.linspace(-1, 1, N_layers)
    X, Y, Z = torch.meshgrid((x, y, z))

    # Load material data
    material_collection = MaterialCollection(
        run_cfg.target_frequencies, to_skip=to_skip
    )

    # Get closest materials
    errors, material_id = _find_closest_material(eps, run_cfg, material_collection)

    cmap = plt.get_cmap("tab20", material_collection.N_materials)

    # Here we plot it
    fig = plt.figure(figsize=(10, 5), dpi=300)
    fig.patch.set_facecolor("white")
    if plot_error:
        ax = fig.add_subplot(121, projection="3d")
    else:
        ax = fig.add_subplot(111, projection="3d")
    ax.view_init(elev=25, azim=100)
    p = ax.scatter(
        X.reshape(-1, 1),
        Y.reshape(-1, 1),
        Z.reshape(-1, 1),
        marker="s",
        s=120,
        linewidths=0,
        cmap=cmap,
        vmin=1 - 0.5,  # This is where we get the discrete colormap
        vmax=material_collection.N_materials
        + 0.5,  # This is where we get the discrete colormap
        alpha=1.0,
        c=material_id.detach().cpu().numpy() + 1,
    )
    cbar = fig.colorbar(
        p, ticks=np.arange(1, material_collection.N_materials + 1)
    )  # This is where we get the discrete colormap
    cbar.set_ticklabels(material_collection.material_names)
    cbar.ax.tick_params(labelsize=NIDN_FONTSIZE - 4)
    # cbar.set_label("Materials", labelpad=-1)

    ax.grid(False)  # Hide grid lines
    # ax.set_xlabel("$N_x =$" + str(Nx))
    # ax.set_ylabel("$N_y =$" + str(Ny))
    # ax.set_zlabel("# of layers", rotation=60)  # TODO Fix rotation
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_zticks(np.linspace(-1, 1, N_layers))

    z_labels = [""] * N_layers
    for idx in range(N_layers):
        z_labels[idx] = "L" + str(idx + 1)
    # Where L1 is (seemingly) the bottom one
    ax.set_zticklabels(z_labels, fontsize=NIDN_FONTSIZE - 4)

    if plot_error:
        ax = fig.add_subplot(122, projection="3d")
        ax.view_init(elev=25, azim=100)
        p = ax.scatter(
            X.reshape(-1, 1),
            Y.reshape(-1, 1),
            Z.reshape(-1, 1),
            marker="s",
            s=120,
            linewidths=0,
            c=errors.detach().cpu().numpy(),
        )
        cbar = plt.colorbar(p, ax=ax)
        cbar.ax.tick_params(labelsize=7)
        ax.grid(False)  # Hide grid lines
        ax.set_xlabel("$N_x =$" + str(Nx))
        ax.set_ylabel("$N_y =$" + str(Ny))
        # ax.set_zlabel("# of layers", rotation=60)  # TODO Fix rotation
        ax.set_xticks([])
        ax.set_yticks([])
        ax.set_zticks(np.linspace(-1, 1, N_layers))

        z_labels = [""] * N_layers
        for idx in range(N_layers):
            z_labels[idx] = "L" + str(idx + 1)
        ax.set_zticklabels(z_labels)  # Where L1 is (seemingly) the bottom one

    if save_path is not None:
        plt.savefig(save_path + "/material_grid.png", dpi=150)



import torch

from nidn.utils.global_constants import PI

from .constants import *

from .utils.fft_funs import Epsilon_fft, get_ifft
from .utils.kbloch import Lattice_Reciprocate, Lattice_getG, Lattice_SetKs
from .utils.torch_functions import torch_transpose, torch_dot, torch_eye, torch_zeros


class TRCWA:
    def __init__(self, nG, L1, L2, freq, theta, phi, verbose=1):
        """Initializes the TRCWA class.

        The time harmonic convention is exp(-i omega t), speed of light = 1.

        Two kinds of layers are currently supported: uniform layer,
        patterned layer from grids. Interface for patterned layer by
        direct analytic expression of Fourier series is included, but
        no examples inclded so far.

        Args:
            nG (int): Truncation order. Note: the actual truncation order might not be nG but smaller.
            L1, L2 (float): Lattice vectors in the x and y direction, respectively. Defines the base unit of the whole system.
            freq (float): The "unitless" frequency, given in relation to the lattice vectors. For a better understanding, read https://web.stanford.edu/group/fan/S4/units.html#length-time-and-frequency.
            theta (float): The polar angle in radians (0 <= theta <= pi).
            phi (float): The azimuthal angle in radians (0 <= theta < 2pi).
            verbose (int, optional): If verbose > 0, the actual nG is printed. Defaults to 1.
        """
        self.freq = freq
        self.omega = 2 * PI * freq + 0.0j
        self.L1 = L1
        self.L2 = L2
        self.phi = torch.tensor(phi)
        self.theta = torch.tensor(theta)
        self.nG = nG
        self.verbose = verbose
        self.Layer_N = 0  # total number of layers

        # the length of the following variables = number of total layers
        self.thickness_list = []
        self.id_list = (
            []
        )  # [type, No., No. in patterned/uniform, No. in its family] starting from 0
        # type:0 for uniform, 1 for Grids, 2 for Fourier

        self.kp_list = []
        self.q_list = []  # eigenvalues
        self.phi_list = []  # eigenvectors

        # Uniform layer
        self.Uniform_ep_list = []
        self.Uniform_N = 0

        # Patterned layer
        self.Patterned_N = 0  # total number of patterned layers
        self.Patterned_epinv_list = []
        self.Patterned_ep2_list = []

        # patterned layer from Grids
        self.GridLayer_N = 0
        self.GridLayer_Nxy_list = []

        # layers of analytic Fourier series (e.g. circles)
        self.FourierLayer_N = 0
        self.FourierLayer_params = []

    def Add_LayerUniform(self, thickness, epsilon):
        # assert type(thickness) == float, 'thickness should be a float'

        self.id_list.append([0, self.Layer_N, self.Uniform_N])
        self.Uniform_ep_list.append(epsilon)
        self.thickness_list.append(torch.tensor(thickness))

        self.Layer_N += 1
        self.Uniform_N += 1

    def Add_LayerGrid(self, thickness, Nx, Ny):
        self.thickness_list.append(thickness)
        self.GridLayer_Nxy_list.append([Nx, Ny])
        self.id_list.append([1, self.Layer_N, self.Patterned_N, self.GridLayer_N])

        self.Layer_N += 1
        self.GridLayer_N += 1
        self.Patterned_N += 1

    def Add_LayerFourier(self, thickness, params):
        self.thickness_list.append(thickness)
        self.FourierLayer_params.append(params)
        self.id_list.append([2, self.Layer_N, self.Patterned_N, self.FourierLayer_N])

        self.Layer_N += 1
        self.Patterned_N += 1
        self.FourierLayer_N += 1

    def Init_Setup(self, Pscale=1.0, Gmethod=0):
        """Set up the reciprocal lattice, compute eigenvalues for uniform layers, and initialize vectors for patterned layers.

        Args:
            Pscale (float, optional): To scale the periodicity in both lateral directions simultaneously (as an autogradable parameter). Period will be Pscale*Lx and Pscale*Ly.
            Gmethod (int, optional): Fourier space truncation scheme; 0 for circular, 1 for rectangular.
        """
        kx0 = (
            self.omega
            * torch.sin(self.theta)
            * torch.cos(self.phi)
            * torch.sqrt(self.Uniform_ep_list[0])
        )
        ky0 = (
            self.omega
            * torch.sin(self.theta)
            * torch.sin(self.phi)
            * torch.sqrt(self.Uniform_ep_list[0])
        )

        # set up reciprocal lattice
        self.Lk1, self.Lk2 = Lattice_Reciprocate(self.L1, self.L2)
        self.G, self.nG = Lattice_getG(self.nG, self.Lk1, self.Lk2, method=Gmethod)

        self.Lk1 = self.Lk1 / Pscale
        self.Lk2 = self.Lk2 / Pscale
        # self.kx = kx0 + 2*torch.pi*(self.Lk1[0]*self.G[:,0]+self.Lk2[0]*self.G[:,1])
        # self.ky = ky0 + 2*torch.pi*(self.Lk1[1]*self.G[:,0]+self.Lk2[1]*self.G[:,1])
        self.kx, self.ky = Lattice_SetKs(self.G, kx0, ky0, self.Lk1, self.Lk2)

        # normalization factor for energies off normal incidence
        self.normalization = torch.sqrt(self.Uniform_ep_list[0]) / torch.cos(self.theta)

        # if comm.rank == 0 and verbose>0:
        if self.verbose > 0:
            print("Total nG = ", self.nG)

        self.Patterned_ep2_list = [None] * self.Patterned_N
        self.Patterned_epinv_list = [None] * self.Patterned_N
        for i in range(self.Layer_N):
            if self.id_list[i][0] == 0:
                ep = self.Uniform_ep_list[self.id_list[i][2]]
                kp = MakeKPMatrix(self.omega, 0, 1.0 / ep, self.kx, self.ky)
                self.kp_list.append(kp)

                q, phi = SolveLayerEigensystem_uniform(self.omega, self.kx, self.ky, ep)
                self.q_list.append(q)
                self.phi_list.append(phi)
            else:
                self.kp_list.append(None)
                self.q_list.append(None)
                self.phi_list.append(None)

    def MakeExcitationPlanewave(
        self, p_amp, p_phase, s_amp, s_phase, order=0, direction="forward"
    ):
        """Sets the excitation to be a planewave incident upon the front (first layer specified) or back of the structure.
        If both tilt angles are specified to be zero, then the planewave is normally incident with the electric field
        polarized along the x-axis for the p-polarization. The phase of each polarization is defined at the origin (z = 0).

        Args:
            p_amp (complex float): The electric field amplitude of the p-polarizations of the planewave. Is a complex number with absolute value between 0 and 1.
            p_phase (float): Phase of the p-polarized part of the light.
            s_amp (complex float): The electric field amplitude of the s-polarizations of the planewave. Is a complex number with absolute value between 0 and 1.
            s_phase (float): Phase of the s-polarized part of the light.
            order (int, optional): A positive integer specifying which order (mode index) to excite. Defaults to 0.
            direction (string, optional): The direction of the planewave (forward is incident upon the front). Defaults to "forward".
        """
        self.direction = direction
        theta = self.theta
        phi = self.phi
        a0 = torch_zeros(2 * self.nG, dtype=complex)
        bN = torch_zeros(2 * self.nG, dtype=complex)
        if direction == "forward":
            tmp1 = torch_zeros(2 * self.nG, dtype=complex)
            tmp1[order] = 1.0
            a0 = a0 + tmp1 * (
                -s_amp
                * torch.cos(theta)
                * torch.cos(phi)
                * torch.exp(torch.tensor(1j * s_phase))
                - p_amp * torch.sin(phi) * torch.exp(torch.tensor(1j * p_phase))
            )

            tmp2 = torch_zeros(2 * self.nG, dtype=complex)
            tmp2[order + self.nG] = 1.0
            a0 = a0 + tmp2 * (
                -s_amp
                * torch.cos(theta)
                * torch.sin(phi)
                * torch.exp(torch.tensor(1j * s_phase))
                + p_amp * torch.cos(phi) * torch.exp(torch.tensor(1j * p_phase))
            )
        elif direction == "backward":
            tmp1 = torch_zeros(2 * self.nG, dtype=complex)
            tmp1[order] = 1.0
            bN = bN + tmp1 * (
                -s_amp
                * torch.cos(theta)
                * torch.cos(phi)
                * torch.exp(torch.tensor(1j * s_phase))
                - p_amp * torch.sin(phi) * torch.exp(torch.tensor(1j * p_phase))
            )

            tmp2 = torch_zeros(2 * self.nG, dtype=complex)
            tmp2[order + self.nG] = 1.0
            bN = bN + tmp2 * (
                -s_amp
                * torch.cos(theta)
                * torch.sin(phi)
                * torch.exp(torch.tensor(1j * s_phase))
                + p_amp * torch.cos(phi) * torch.exp(torch.tensor(1j * p_phase))
            )

        self.a0 = a0
        self.bN = bN

    def GridLayer_geteps(self, ep_all):
        """Fourier transform + eigenvalue for grid layer.

        Args:
            ep_all (torch.tensor): A tensor containing all epsilon values.
        """
        ptri = 0
        ptr = 0
        for i in range(self.Layer_N):
            if self.id_list[i][0] != 1:
                continue

            Nx = self.GridLayer_Nxy_list[ptri][0]
            Ny = self.GridLayer_Nxy_list[ptri][1]
            dN = 1.0 / Nx / Ny

            if len(ep_all) == 3 and ep_all[0].ndim > 0:
                ep_grid = [
                    torch.reshape(ep_all[0][ptr : ptr + Nx * Ny], [Nx, Ny]),
                    torch.reshape(ep_all[1][ptr : ptr + Nx * Ny], [Nx, Ny]),
                    torch.reshape(ep_all[2][ptr : ptr + Nx * Ny], [Nx, Ny]),
                ]
            elif len(ep_all.shape) == 3:
                ep_grid = ep_all[:, :, i - 1]
            else:
                ep_grid = torch.reshape(
                    torch.tensor(ep_all[ptr : ptr + Nx * Ny]), [Nx, Ny]
                )

            epinv, ep2 = Epsilon_fft(dN, ep_grid, self.G)

            self.Patterned_epinv_list[self.id_list[i][2]] = epinv
            self.Patterned_ep2_list[self.id_list[i][2]] = ep2

            kp = MakeKPMatrix(self.omega, 1, epinv, self.kx, self.ky)
            self.kp_list[self.id_list[i][1]] = kp

            q, phi = SolveLayerEigensystem(self.omega, self.kx, self.ky, kp, ep2)
            self.q_list[self.id_list[i][1]] = q
            self.phi_list[self.id_list[i][1]] = phi

            ptr += Nx * Ny
            ptri += 1

    def Return_eps(self, which_layer, Nx, Ny, component="xx"):
        """Used to get real-space epsilon profile reconstructured from the truncated Fourier orders.

        Args:
            which_layer (int): Which layer to return the epsilon values from.
            Nx (int): Grid points within the unit cell in the x direction.
            Ny (int): Grid points within the unit cell in the y direction.
            component (string, optional): For patterned layer, component = 'xx','xy','yx','yy','zz'. For uniform layer, currently it's assumed to be isotropic. Defaults to "xx".

        Returns:
            torch.tensor: Real-space epsilon profile
        """
        i = which_layer
        # uniform layer
        if self.id_list[i][0] == 0:
            ep = self.Uniform_ep_list[self.id_list[i][2]]
            return torch.ones((Nx, Ny)) * ep

        # patterned layer
        elif self.id_list[i][0] == 1:
            if component == "zz":
                epk = torch.inv(self.Patterned_epinv_list[self.id_list[i][2]])
            elif component == "xx":
                epk = self.Patterned_ep2_list[self.id_list[i][2]][: self.nG, : self.nG]
            elif component == "xy":
                epk = self.Patterned_ep2_list[self.id_list[i][2]][: self.nG, self.nG :]
            elif component == "yx":
                epk = self.Patterned_ep2_list[self.id_list[i][2]][self.nG :, : self.nG]
            elif component == "yy":
                epk = self.Patterned_ep2_list[self.id_list[i][2]][self.nG :, self.nG :]

            return get_ifft(Nx, Ny, epk[0, :], self.G)

    def RT_Solve(self, normalize=0, byorder=0):
        """
        Reflection and transmission power computation.
        Returns 2R and 2T, following Victor Liu's notation (https://web.stanford.edu/group/fan/S4/python_api.html?highlight=power#S4.Simulation.GetPowerFlux).

        Args:
            normalize (int, optional): To normalize the output when the 0-th media is not vacuum, or for oblique incidence. If normalize = 1, the output will be divided by n[0]*cos(theta). Defaults to 0.
            byorder (int, optional): To get Poynting flux by order. If 0, the total power is computed. Defaults to 0.

        Returns:
            torch.tensor: Reflection power, Transmission power
        """
        aN, b0 = SolveExterior(
            self.a0,
            self.bN,
            self.q_list,
            self.phi_list,
            self.kp_list,
            self.thickness_list,
        )

        fi, bi = GetZPoyntingFlux(
            self.a0,
            b0,
            self.omega,
            self.kp_list[0],
            self.phi_list[0],
            self.q_list[0],
            byorder=byorder,
        )

        fe, be = GetZPoyntingFlux(
            aN,
            self.bN,
            self.omega,
            self.kp_list[-1],
            self.phi_list[-1],
            self.q_list[-1],
            byorder=byorder,
        )

        if self.direction == "forward":
            R = torch.real(-bi)
            T = torch.real(fe)
        elif self.direction == "backward":
            R = torch.real(fe)
            T = torch.real(-bi)

        if normalize == 1:
            R = R * self.normalization
            T = T * self.normalization
        return R, T

    def GetAmplitudes_noTranslate(self, which_layer):
        """To get the amplitude of the eigenvectors at some layer.

        Args:
            which_layer (int): Which layer to return the Fourier amplitude from.

        Returns:
            torch.tensor: ai, bi; each containing the complex amplitudes of each forward and backward mode
        """
        if which_layer == 0:
            aN, b0 = SolveExterior(
                self.a0,
                self.bN,
                self.q_list,
                self.phi_list,
                self.kp_list,
                self.thickness_list,
            )
            ai = self.a0
            bi = b0

        elif which_layer == self.Layer_N - 1:
            aN, b0 = SolveExterior(
                self.a0,
                self.bN,
                self.q_list,
                self.phi_list,
                self.kp_list,
                self.thickness_list,
            )
            ai = aN
            bi = self.bN

        else:
            ai, bi = SolveInterior(
                which_layer,
                self.a0,
                self.bN,
                self.q_list,
                self.phi_list,
                self.kp_list,
                self.thickness_list,
            )
        return ai, bi

    def GetAmplitudes(self, which_layer, z_offset):
        """To get the Fourier amplitude of the eigenvectors at some layer and at some zoffset.
        Returns the raw mode amplitudes within a particular layer. For uniform (unpatterned) layers,
        the modes are simply the diffracted orders The first value is guaranteed to be the straight transmitted or specularly reflected diffraction order. For patterned layers, there is typically no meaningful information in these amplitudes.

        Args:
            which_layer (int): Which layer to return the Fourier amplitude from.
            z_offset (float): The z-offset at which to obtain the mode amplitudes. Must be 0 < z_offset < layer thickness.

        Returns:
            torch.tensor: ai, bi; containing the complex amplitudes of each forward and backward mode
        """
        if which_layer == 0:
            aN, b0 = SolveExterior(
                self.a0,
                self.bN,
                self.q_list,
                self.phi_list,
                self.kp_list,
                self.thickness_list,
            )
            ai = self.a0
            bi = b0

        elif which_layer == self.Layer_N - 1:
            aN, b0 = SolveExterior(
                self.a0,
                self.bN,
                self.q_list,
                self.phi_list,
                self.kp_list,
                self.thickness_list,
            )
            ai = aN
            bi = self.bN

        else:
            ai, bi = SolveInterior(
                which_layer,
                self.a0,
                self.bN,
                self.q_list,
                self.phi_list,
                self.kp_list,
                self.thickness_list,
            )

        ai, bi = TranslateAmplitudes(
            self.q_list[which_layer], self.thickness_list[which_layer], z_offset, ai, bi
        )

        return ai, bi

    def Solve_FieldFourier(self, which_layer, z_offset):
        """Returns the field amplitude in Fourier space: [Ex,Ey,Ez], [Hx,Hy,Hz]
        Args:
            which_layer (int): Which layer to return the field amplitude from.
            z_offset (torch.tensor): The z-offset at which to obtain the field amplitude. Must be 0 < z_offset < layer thickness. Can be a single number or a list of numbers.

        Returns:
            torch.tensor: Tensors containing the field amplitudes [Ex,Ey,Ez], [Hx,Hy,Hz] in Fourier space
        """
        ai0, bi0 = self.GetAmplitudes_noTranslate(which_layer)
        # ai, bi = self.GetAmplitudes(which_layer,z_offset)

        if torch.isinstance(z_offset, float) or torch.isinstance(z_offset, int):
            zl = [z_offset]
        else:
            zl = z_offset

        eh = []
        for zoff in zl:
            ai, bi = TranslateAmplitudes(
                self.q_list[which_layer],
                self.thickness_list[which_layer],
                zoff,
                ai0,
                bi0,
            )
            # hx, hy in Fourier space
            fhxy = torch_dot(self.phi_list[which_layer], ai + bi)
            fhx = fhxy[: self.nG]
            fhy = fhxy[self.nG :]

            # ex,ey in Fourier space
            tmp1 = (ai - bi) / self.omega / self.q_list[which_layer]
            tmp2 = torch_dot(self.phi_list[which_layer], tmp1)
            fexy = torch_dot(self.kp_list[which_layer], tmp2)
            fey = -fexy[: self.nG]
            fex = fexy[self.nG :]

            # hz in Fourier space
            fhz = (self.kx * fey - self.ky * fex) / self.omega

            # ez in Fourier space
            fez = (self.ky * fhx - self.kx * fhy) / self.omega
            if self.id_list[which_layer][0] == 0:
                fez = fez / self.Uniform_ep_list[self.id_list[which_layer][2]]
            else:
                fez = torch_dot(
                    self.Patterned_epinv_list[self.id_list[which_layer][2]], fez
                )
            eh.append([[fex, fey, fez], [fhx, fhy, fhz]])
        return eh

    def Solve_FieldOnGrid(self, which_layer, z_offset, Nxy=None):
        """To get fields in real space on grid points. If single z_offset, the output is [[Ex,Ey,Ez], [Hx,Hy,Hz]].
        If z_offset is a list, the output os [[[Ex1,Ey1,Ez1],[Hx1,Hy1,Hz1]],  [[Ex2,Ey2,Ez2],[Hx2,Hy2,Hz2]],  ...].

        Args:
            which_layer (int): Which layer to return the field amplitude from.
            z_offset (torch.tensor): The z-offset at which to obtain the field amplitude. Must be 0 < z_offset < layer thickness. Can be a single number or a list of numbers.
            Nxy (torch.tensor, optional): Nxy = [Nx,Ny], if not supplied, will use the number in patterned layer.

        Returns:
            torch.tensor: Tensor containing the field amplitudes [[Ex,Ey,Ez], [Hx,Hy,Hz]] in Fourier space for each z-offset
        """

        if torch.isinstance(Nxy, type(None)):
            Nxy = self.GridLayer_Nxy_list[self.id_list[which_layer][3]]
        Nx = Nxy[0]
        Ny = Nxy[1]

        # e,h in Fourier space
        fehl = self.Solve_FieldFourier(which_layer, z_offset)

        eh = []
        for feh in fehl:
            fe = feh[0]
            fh = feh[1]
            ex = get_ifft(Nx, Ny, fe[0], self.G)
            ey = get_ifft(Nx, Ny, fe[1], self.G)
            ez = get_ifft(Nx, Ny, fe[2], self.G)

            hx = get_ifft(Nx, Ny, fh[0], self.G)
            hy = get_ifft(Nx, Ny, fh[1], self.G)
            hz = get_ifft(Nx, Ny, fh[2], self.G)
            eh.append([[ex, ey, ez], [hx, hy, hz]])
        if torch.isinstance(z_offset, float) or torch.isinstance(z_offset, int):
            eh = eh[0]
        return eh

    def Volume_integral(self, which_layer, Mx, My, Mz, normalize=0):
        """To get volume integration with respect to some convolution matrix M defined for 3 directions, respectively.
        Returns the volume integral of a particular density over a unit cell throughout the entire thickness of a layer.
        Mxyz is the convolution matrix. This function computes 1/A\int_V Mx|Ex|^2+My|Ey|^2+Mz|Ez|^2.
        To be consistent with Poynting vector defintion here, the absorbed power will be just omega*output.

        Args:
            which_layer (int): Which layer to get the volume integral from.
            Mx (): Some convolution matrix M defined for the x-direction.
            My (): Some convolution matrix M defined for the y-direction.
            Mz (): Some convolution matrix M defined for the z-direction.
            normalize (int, optional): If 1, the return value is normalized according to self.normalization. Defaults to 0.

        Returns:
            float: Integration value.
        """
        kp = self.kp_list[which_layer]
        q = self.q_list[which_layer]
        phi = self.phi_list[which_layer]

        if self.id_list[which_layer][0] == 0:
            epinv = 1.0 / self.Uniform_ep_list[self.id_list[which_layer][2]]
        else:
            epinv = self.Patterned_epinv_list[self.id_list[which_layer][2]]

        # un-translated amplitude
        ai, bi = SolveInterior(
            which_layer,
            self.a0,
            self.bN,
            self.q_list,
            self.phi_list,
            self.kp_list,
            self.thickness_list,
        )
        ab = torch.hstack((ai, bi))
        abMatrix = torch.outer(ab, torch.conj(ab))

        Mt = Matrix_zintegral(q, self.thickness_list[which_layer])
        # overall
        abM = abMatrix * Mt

        # F matrix
        Faxy = torch_dot(torch_dot(kp, phi), torch.diag(1.0 / self.omega / q))
        Faz1 = 1.0 / self.omega * torch_dot(epinv, torch.diag(self.ky))
        Faz2 = -1.0 / self.omega * torch_dot(epinv, torch.diag(self.kx))
        Faz = torch_dot(torch.hstack((Faz1, Faz2)), phi)

        tmp1 = torch.vstack((Faxy, Faz))
        tmp2 = torch.vstack((-Faxy, Faz))
        F = torch.hstack((tmp1, tmp2))

        # consider Mtotal
        Mzeros = torch.zeros_like(Mx)
        Mtotal = torch.vstack(
            (
                torch.hstack((Mx, Mzeros, Mzeros)),
                torch.hstack((Mzeros, My, Mzeros)),
                torch.hstack((Mzeros, Mzeros, Mz)),
            )
        )

        # integral = Tr[ abMatrix * F^\dagger *  Matconv *F ]
        tmp = torch_dot(torch_dot(torch.conj(torch_transpose(F)), Mtotal), F)
        val = torch.trace(torch_dot(abM, tmp))

        if normalize == 1:
            val = val * self.normalization
        return val

    def Solve_ZStressTensorIntegral(self, which_layer):
        """To compute the Maxwell stress tensor, integrated over the z-plane.
        Returns the integral of the electromagnetic stress tensor over a unit cell
        surface normal to the z-direction.
        Returns 2F_x,2F_y,2F_z, integrated over the z-plane.

        Args:
            which_layer (int): The layer in which the integration surface lies.

        Returns:
            torch.tensor: The real and imaginary parts of the x-, y-, and z-components of the stress tensor integrated over the specified surface, assuming a unit normal vector in the +z direction.
        """
        z_offset = 0.0
        eh = self.Solve_FieldFourier(which_layer, z_offset)
        e = eh[0][0]
        h = eh[0][1]
        ex = e[0]
        ey = e[1]
        ez = e[2]

        hx = h[0]
        hy = h[1]
        hz = h[2]

        # compute D = epsilon E
        ## Dz = epsilon_z E_z = (ky*hx - kx*hy)/self.omega
        dz = (self.ky * hx - self.kx * hy) / self.omega

        ## Dxy = epsilon2 * Exy
        if self.id_list[which_layer][0] == 0:
            dx = ex * self.Uniform_ep_list[self.id_list[which_layer][2]]
            dy = ey * self.Uniform_ep_list[self.id_list[which_layer][2]]
        else:
            exy = torch.hstack((-ey, ex))
            dxy = torch_dot(self.Patterned_ep2_list[self.id_list[which_layer][2]], exy)
            dx = dxy[self.nG :]
            dy = -dxy[: self.nG]

        Tx = torch.sum(ex * torch.conj(dz) + hx * torch.conj(hz))
        Ty = torch.sum(ey * torch.conj(dz) + hy * torch.conj(hz))
        Tz = 0.5 * torch.sum(
            ez * torch.conj(dz)
            + hz * torch.conj(hz)
            - ey * torch.conj(dy)
            - ex * torch.conj(dx)
            - torch.abs(hx) ** 2
            - torch.abs(hy) ** 2
        )

        Tx = torch.real(Tx)
        Ty = torch.real(Ty)
        Tz = torch.real(Tz)

        return Tx, Ty, Tz


def MakeKPMatrix(omega, layer_type, epinv, kx, ky):
    """Makes KP matrix.

    Args:
        omega (float): The angular frequency.
        layer_type (int): Uniform or patterned layer. Set to 0 for uniform layers and >0 for patterned.
        epinv ():
        kx ():
        ky ():

    Returns:
        torch.tensor:
    """
    nG = len(kx)

    # uniform layer, epinv has length 1
    if layer_type == 0:
        # JkkJT = np.block([[np.diag(ky*ky), np.diag(-ky*kx)],
        #                 [np.diag(-kx*ky),np.diag(kx*kx)]])

        Jk = torch.vstack((torch.diag(-ky), torch.diag(kx)))
        JkkJT = torch_dot(Jk, torch_transpose(Jk))

        kp = omega**2 * torch_eye(2 * nG) - epinv * JkkJT
    # patterned layer
    else:
        Jk = torch.vstack((torch.diag(-ky), torch.diag(kx)))
        tmp = torch_dot(Jk, epinv)
        kp = omega**2 * torch_eye(2 * nG) - torch_dot(tmp, torch_transpose(Jk))

    return kp


def SolveLayerEigensystem_uniform(omega, kx, ky, epsilon):
    """Solves the eigensystem for uniform layers.

    Args:
        omega (float): The angular frequency.
        kx ():
        ky ():
        epsilon ():

    Returns:
        torch.tensor, torch.tensor: q, phi
    """
    nG = len(kx)
    q = torch.sqrt(epsilon * omega**2 - kx**2 - ky**2)
    # branch cut choice
    q = torch.where(torch.imag(q) < 0.0, -q, q)

    q = torch.cat((q, q))
    phi = torch_eye(2 * nG)
    return q, phi


def SolveLayerEigensystem(omega, kx, ky, kp, ep2):
    """Solves the eigensystem.

    Args:
        omega (float): The angular frequency. Unused (from GRCWA).
        kx ():
        ky ():
        kp ():
        ep2 ():

    Returns:
        torch.tensor, torch.tensor: q, phi
    """
    nG = len(kx)

    k = torch.vstack((torch.diag(kx), torch.diag(ky)))
    kkT = torch_dot(k, torch_transpose(k))
    M = torch_dot(ep2, kp) - kkT

    q, phi = torch.linalg.eig(M)

    q = torch.sqrt(q)
    # branch cut choice
    q = torch.where(torch.imag(q) < 0.0, -q, q)
    return q, phi


def GetSMatrix(indi, indj, q_list, phi_list, kp_list, thickness_list):
    """Gets the S matrices, S_ij: size 4n*4n.

    Args:
        indi (int): Start layer index.
        indj (int): End layer index. Must be >= indi.
        q_list ():
        phi_list ():
        kp_list ():
        thickness_list (): List of thicknesses for each layer in the stack.

    Returns:
        torch.tensor, torch.tensor, torch.tensor, torch.tensor: S11, S12, S21, S22
    """
    # assert type(indi) == int, 'layer index i must be integer'
    # assert type(indj) == int, 'layer index j must be integer'

    nG2 = len(q_list[0])
    S11 = torch_eye(nG2, dtype=complex)
    S12 = torch.zeros_like(S11)
    S21 = torch.zeros_like(S11)
    S22 = torch_eye(nG2, dtype=complex)
    if indi == indj:
        return S11, S12, S21, S22
    elif indi > indj:
        raise Exception("indi must be < indj")

    for l in range(indi, indj):
        ## next layer
        lp1 = l + 1
        ## Q = inv(phi_l) * phi_lp1
        Q = torch_dot(torch.inverse(phi_list[l]), phi_list[lp1])
        ## P = ql*inv(kp_l*phi_l) * kp_lp1*phi_lp1*q_lp1^-1
        P1 = torch_dot(
            torch.diag(q_list[l]), torch.inverse(torch_dot(kp_list[l], phi_list[l]))
        )
        P2 = torch_dot(
            torch_dot(kp_list[lp1], phi_list[lp1]), torch.diag(1.0 / q_list[lp1])
        )
        P = torch_dot(P1, P2)
        # P1 = torch_dot(kp_list[l],phi_list[l])
        # P2 = torch_dot(torch_dot(kp_list[lp1],phi_list[lp1]),   torch.diag(1./q_list[lp1]))
        # P = np.linalg.solve(P1,P2)
        # P = np.dot(np.diag(q_list[l]),P)

        # T11=T22, T12=T21
        T11 = 0.5 * (Q + P)
        T12 = 0.5 * (Q - P)

        # phase
        d1 = torch.diag(torch.exp(1j * q_list[l] * thickness_list[l]))
        d2 = torch.diag(torch.exp(1j * q_list[lp1] * thickness_list[lp1]))

        # S11 = inv(T11-d1*S12o*T12)*d1*S11o
        P1 = T11 - torch_dot(torch_dot(d1, S12), T12)
        P1 = torch.inverse(P1)  # hold for further use
        S11 = torch_dot(torch_dot(P1, d1), S11)

        # S12 = P1*(d1*S12o*T11-T12)*d2
        P2 = torch_dot(d1, torch_dot(S12, T11)) - T12
        S12 = torch_dot(torch_dot(P1, P2), d2)

        # S21 = S22o*T12*S11+S21o
        S21 = S21 + torch_dot(S22, torch_dot(T12, S11))

        # S22 = S22o*T12*S12+S22o*T11*d2
        P2 = torch_dot(S22, torch_dot(T12, S12))
        P1 = torch_dot(S22, torch_dot(T11, d2))
        S22 = P1 + P2

    return S11, S12, S21, S22


def SolveExterior(a0, bN, q_list, phi_list, kp_list, thickness_list):
    """Given a0, bN, solve for b0, aN.

    Args:
        a0 ():
        bN ():
        q_list ():
        phi_list ():
        kp_list ():
        thickness_list (): List of thicknesses for each layer in the stack.

    Returns:
        torch.tensor, torch.tensor: aN, b0
    """

    Nlayer = len(thickness_list)  # total number of layers
    S11, S12, S21, S22 = GetSMatrix(
        0, Nlayer - 1, q_list, phi_list, kp_list, thickness_list
    )

    aN = torch_dot(S11, a0) + torch_dot(S12, bN)
    b0 = torch_dot(S21, a0) + torch_dot(S22, bN)

    return aN, b0


def SolveInterior(which_layer, a0, bN, q_list, phi_list, kp_list, thickness_list):
    """
    Given a0, bN, solve for ai, bi

    Args:
        which_layer (int): The layer to solve for. Layer numbering starts from 0.
        a0 ():
        bN ():
        q_list ():
        phi_list ():
        kp_list ():
        thickness_list (): List of thicknesses for each layer in the stack.

    Returns:
        torch.tensor, torch.tensor: ai, bi
    """
    Nlayer = len(thickness_list)  # total number of layers
    nG2 = len(q_list[0])

    S11, S12, S21, S22 = GetSMatrix(
        0, which_layer, q_list, phi_list, kp_list, thickness_list
    )
    pS11, pS12, pS21, pS22 = GetSMatrix(
        which_layer, Nlayer - 1, q_list, phi_list, kp_list, thickness_list
    )

    # tmp = inv(1-S12*pS21)
    tmp = torch.inv(torch_eye(nG2) - torch_dot(S12, pS21))
    # ai = tmp * (S11 a0 + S12 pS22 bN)
    ai = torch_dot(tmp, torch_dot(S11, a0) + torch_dot(S12, torch_dot(pS22, bN)))
    # bi = pS21 ai + pS22 bN
    bi = torch_dot(pS21, ai) + torch_dot(pS22, bN)

    return ai, bi


def TranslateAmplitudes(q, thickness, dz, ai, bi):
    """

    Args:
        q (): q for the layer of interest.
        thickness (float): The thickness of the layer of interest.
        dz (float): The z-offset at which to translate the amplitudes.
        ai ():
        bi ():

    Returns:
        torch.tensor, torch.tensor: aim, bim
    """
    aim = ai * torch.exp(1j * q * dz)
    bim = bi * torch.exp(1j * q * (thickness - dz))
    return aim, bim


def GetZPoyntingFlux(ai, bi, omega, kp, phi, q, byorder=0):
    """Returns 2S_z/A, following Victor Liu's notation.

    Args:
        ai ():
        bi ():
        omega (float): The angular frequency.
        kp ():
        phi ():
        q ():
        byorder (int, optional): If the Poynting flux is to be given for each order (byorder > 0) or summed (= 0). Defaults to 0.

    Returns:
        torch.tensor, torch.tensor: The Poynting flux in the forward and backward direction, given as a number or a list of numbers
    """
    n2 = len(ai)
    n = int(n2 / 2)
    # A = kp phi inv(omega*q)
    A = torch_dot(torch_dot(kp, phi), torch.diag(1.0 / omega / q))

    pa = torch_dot(phi, ai)
    pb = torch_dot(phi, bi)
    Aa = torch_dot(A, ai)
    Ab = torch_dot(A, bi)

    # diff = 0.5*(pb* Aa - Ab* pa)
    diff = 0.5 * (torch.conj(pb) * Aa - torch.conj(Ab) * pa)
    # forward = real(Aa* pa) + diff
    forward_xy = torch.real(torch.conj(Aa) * pa) + diff
    backward_xy = -torch.real(torch.conj(Ab) * pb) + torch.conj(diff)

    forward = forward_xy[:n] + forward_xy[n:]
    backward = backward_xy[:n] + backward_xy[n:]
    if byorder == 0:
        forward = torch.sum(forward)
        backward = torch.sum(backward)

    return forward, backward


def Matrix_zintegral(q, thickness, shift=1e-12):
    """Generate matrix for z-integral.

    Args:
        q ():
        thickness (float): Thickness of the layer of interest.
        shift (float, optional):               . Defaults to 1e-12.

    Returns:
        torch.tensor: Matrix for z-integral
    """
    nG2 = len(q)
    qi, qj = Gmeshgrid(q)

    # # Maa = \int exp(i q_i z)^* exp(i q_j z)
    # #     = [exp(i(q_j-q_i^*)t)-1]/i(q_j-q_i^*)
    # # Mbb = \int exp(i q_i (t-z))^* exp(i q_j (t-z))
    # #     = exp(i(q_j-q_i^*)t)  [exp(i(q_i^*-q_j)t)-1]/i(q_i^*-q_j)
    # #     = Maa
    # # Mab = \int exp(i q_i z)^* exp(i q_j (t-z))
    # #     = exp(iq_j t) [1-exp(-i(q_i^*+q_j)t)]/i(q_i^*+q_j)
    # #     = [exp(iq_j t)-exp(-i q_i^* t)]/i(q_i^*+q_j)
    # # Mba = \int exp(i q_i (t-z))^* exp(i q_j z)
    # #     = exp(-iq_i^* t) [exp(i(q_j+q_i^*)t)-1]/i(q_j+q_i^*)
    # #     = Mab
    # Maa = (np.exp(1j*(qj-np.conj(qi))*thickness)-1)/1j/(qj-np.conj(qi))
    # Mab = (np.exp(1j*qj*thickness)-np.exp(-1j*np.conj(qi)*thickness))/1j/(qj+np.conj(qi))

    # M = t exp(0.5it (qj-qi^*)) sinc(0.5d (sjqj-siqi^*), note in python sinc = sin(pi x)/pi x
    qij = qj - torch.conj(qi) + torch_eye(nG2) * shift
    Maa = (torch.exp(1j * qij * thickness) - 1) / 1j / qij  # this is more robust
    # Maa = thickness * np.exp(0.5j*thickness*qij) * np.sinc(0.5*thickness*qij/np.pi)

    qij2 = qj + torch.conj(qi)
    # Mab = thickness * np.exp(0.5j*thickness*qij) * np.sinc(0.5*thickness*qij2/np.pi)
    Mab = (
        (torch.exp(1j * qj * thickness) - torch.exp(-1j * torch.conj(qi) * thickness))
        / 1j
        / qij2
    )

    tmp1 = torch.vstack((Maa, Mab))
    tmp2 = torch.vstack((Mab, Maa))
    Mt = torch.hstack((tmp1, tmp2))
    return Mt


def Gmeshgrid(x):
    """.

    Args:
        x ():

    Returns:
        torch.tensor, torch.tensor: qi, qj
    """
    N = len(x)
    qj = []
    for i in range(N):
        qj.append(x)
    qj = torch.tensor(qj)
    qi = torch_transpose(qj)
    return qi, qj



from torch import nn
import torch


class AbsLayer(nn.Module):
    """Very simple activation layer to allow different abs layer activations of the siren."""

    def __init__(self):
        super().__init__()

    def forward(self, input):
        return torch.abs(input)



from matplotlib import pyplot as plt
import numpy as np


def plot_losses(run_cfg, save_path=None):
    """Plots the loss over training, optionally saving it.

    Args:
        run_cfg (DotMap): Run configuration.
        save_path (str, optional): Folder to save the plot in. Defaults to None, then the plot will not be saved.
    """
    fig = plt.figure(figsize=(8, 4), dpi=150)
    fig.patch.set_facecolor("white")
    plt.semilogy(run_cfg.results.L1_errs)
    plt.semilogy(run_cfg.results.loss_log)
    plt.semilogy(run_cfg.results.weighted_average_log)
    plt.xlabel("Model evaluations")
    plt.ylabel("Loss")
    plt.legend(["L1", "Loss", "Weighted Average Loss"])
    if save_path is not None:
        plt.savefig(save_path + "/losses.png", dpi=150)



""" often used constants go here """

FDTD_X: int = 0
""" x-index (useful for indexing) """

FDTD_Y: int = 1
""" y-index (useful for indexing) """

FDTD_Z: int = 2
""" z-index (useful for indexing) """



from loguru import logger


from ..fdtd_integration.compute_spectrum_fdtd import compute_spectrum_fdtd
from ..trcwa.compute_spectrum_trcwa import compute_spectrum_trcwa
from .validate_config import _validate_config


def compute_spectrum(eps_grid, cfg, validate_cfg=True):
    """Computes spectrum, using either TRCWA or FDTD simulations

    Args:
        cfg (DotMap): Configurations for the simulation
        permittivity (Array[Nx][Ny][Nlayers][Nfrequencies]): Matrix of permittivity tensors
        validate_cfg (bool): Whether to validate the config

    Returns:
        Array, Array: Reflection spectrumand transmission spectrum
    """
    if validate_cfg:
        _validate_config(cfg)
    if cfg.solver == "FDTD":
        logger.debug("Using FDTD solver to find spectrum")
        return compute_spectrum_fdtd(eps_grid, cfg)
    elif cfg.solver == "TRCWA":
        logger.debug("Using TRCWA solver to find spectrum")
        return compute_spectrum_trcwa(eps_grid, cfg)



from loguru import logger
from pathlib import Path

from .plot_eps_per_point import plot_eps_per_point
from .plot_losses import plot_losses
from .plot_material_grid import plot_material_grid
from .plot_model_grid import plot_model_grid
from .plot_model_grid_per_freq import plot_model_grid_per_freq
from .plot_spectra import plot_spectra

def save_all_plots(run_cfg, save_path):
    """Creates all plots for the passed run_config and save them in the passed folder.

    Args:
        run_cfg (DotMap): Run configuration (incl. trained model).
        save_path (str): Folder to save the plots to.
    """
    logger.info("Saving all plots to {}".format(save_path))

    # Create subfolder if it does not exist
    Path(save_path).mkdir(parents=True, exist_ok=True)

    logger.debug("Saving spectra comparison plot")
    plot_spectra(run_cfg, save_path=save_path)

    logger.debug("Saving losses plot")
    plot_losses(run_cfg, save_path=save_path)

    logger.debug("Saving model grid plot")
    plot_model_grid(run_cfg, save_path=save_path)

    logger.debug("Saving model grid per freq plot")
    plot_model_grid_per_freq(run_cfg, save_path=save_path)

    logger.debug("Saving material grid plot")
    plot_material_grid(run_cfg, save_path=save_path)

    logger.debug("Saving eps per point plot")
    plot_eps_per_point(run_cfg, save_path=save_path)



import torch
from dotmap import DotMap


def _get_transformed_grid(Nx_undersampled, Ny_undersampled, N_layers, thicknesses):
    """Creates x,y,z in the space [-1,1]**3 . Z weighted by thicknesses

    Args:
          model (torch.model): Trained neural network model. Should map one 4D input.
        Nx_undersampled (int): Number of grid points in x direction. Potentially unesampled if eps_oversampling > 1.
        Ny_undersampled (int): Number of grid points in y direction. Potentially unesampled if eps_oversampling > 1.
        N_layers (int): Number of layers in the model.
        target_frequencies (list): Target frequencies.
        freq_distribution (list): Distribution of points in frequency space. Should be "linear" or "log".
        thicknesses (list): List of thicknesses for each layer or scalar for all.
    Returns:
       [torch.tensor,torch.tensor,torch.tensor]: x,y,z

    """
    # Get the grid ticks
    x = torch.linspace(-1, 1, Nx_undersampled)
    y = torch.linspace(-1, 1, Ny_undersampled)

    # Space z uniform for equally thick layers, else weight by thickness
    if len(thicknesses) == 1:
        z = torch.linspace(-1, 1, N_layers)
    else:
        total_thickness = sum(thicknesses)

        # create weighted z
        z = torch.cumsum(torch.tensor(thicknesses) / total_thickness, dim=0)
        z -= thicknesses[0] / total_thickness

        z = z / z.max() - 0.5
        z *= 2

    return x, y, z


def _avoid_zero_eps(eps, cutoff=1e-2):
    """Clips epsilon values close to zero to either cutoff or -cutoff.

    Args:
        eps (torch.tensor): epsilon to be clipped.
        cutoff (float, optional): Cutoff border. Defaults to 1e-2.

    Returns:
        torch.tensor: modified eps tensor
    """
    # catch zeroes
    indices = eps.real == 0.0
    eps.real[indices] = cutoff
    # catch close to zeroes
    indices = torch.logical_and(eps.real < cutoff, eps.real > -cutoff)
    eps.real[indices] = cutoff * torch.sign(eps.real[indices])
    return eps


def _eval_model(
    model,
    Nx_undersampled,
    Ny_undersampled,
    N_layers,
    target_frequencies,
    freq_distribution,
    thicknesses,
):
    """Evaluates the model on the grid.

    Args:
        model (torch.model): Trained neural network model. Should map one 4D input.
        Nx_undersampled (int): Number of grid points in x direction. Potentially unesampled if eps_oversampling > 1.
        Ny_undersampled (int): Number of grid points in y direction. Potentially unesampled if eps_oversampling > 1.
        N_layers (int): Number of layers in the model.
        target_frequencies (list): Target frequencies.
        freq_distribution (list): Distribution of points in frequency space. Should be "linear" or "log".
        thicknesses (list): List of thicknesses for each layer or scalar for all.
    Returns:
       [torch.tensor]: Resulting 4D [real,imag] epsilon grid
    """
    x, y, z = _get_transformed_grid(
        Nx_undersampled, Ny_undersampled, N_layers, thicknesses
    )

    # Scales sampling domain of frequencies
    freq_scaling = 32.0

    # Layers should be independent, so we scale the sampling in that dimension
    z_scaling = 1000.0
    z *= z_scaling

    if freq_distribution == "linear":
        # Linearly spaced frequency points
        freq = torch.linspace(-freq_scaling, freq_scaling, len(target_frequencies))
    elif freq_distribution == "log":
        # Logspaced frequency points as in cfg
        # Normalize to max val = 1
        freq = torch.tensor(target_frequencies / max(target_frequencies))
        # Transform to -scaling to scaling
        freq = (freq - 0.5) * 2 * freq_scaling
    else:
        raise ValueError("Unknown frequency distribution. Should be 'log' or 'linear'")

    # Create a meshgrid from the grid ticks
    X, Y, Z, FREQ = torch.meshgrid((x, y, z, freq))

    # Reshape the grid to have a Nx4 tensor
    nn_inputs = torch.cat(
        (X.reshape(-1, 1), Y.reshape(-1, 1), Z.reshape(-1, 1), FREQ.reshape(-1, 1)),
        dim=1,
    )

    # Compute model output at each grid point
    return model(nn_inputs)


def _regression_model_to_eps_grid(model, run_cfg: DotMap):
    """Computes a 4D grid of epsilons for a regression model.

    Args:
        model (torch.model): Trained neural network model. Should map one 4D input to a [real,imag] epsilon.
        run_cfg (DotMap): Configuration for the run.

    Returns:
        [torch.tensor]: Resulting 4D [real,imag] epsilon grid
    """

    Nx_undersampled = run_cfg.Nx // run_cfg.eps_oversampling
    Ny_undersampled = run_cfg.Ny // run_cfg.eps_oversampling

    # Evaluate the model at the grid points
    out = _eval_model(
        model,
        Nx_undersampled,
        Ny_undersampled,
        run_cfg.N_layers,
        run_cfg.target_frequencies,
        run_cfg.freq_distribution,
        run_cfg.PER_LAYER_THICKNESS,
    )

    # Reshape the output to have a 4D tensor again
    # Note we ouput real and imaginary parts separatly, hence 2*N_freq
    out = out.reshape(
        Nx_undersampled, Ny_undersampled, run_cfg.N_layers, 2 * run_cfg.N_freq
    )

    # Oversample the created eps grid
    out = torch.repeat_interleave(out, run_cfg.eps_oversampling, dim=0)
    out = torch.repeat_interleave(out, run_cfg.eps_oversampling, dim=1)

    # Initialize the epsilon grid
    eps = torch.zeros(
        [run_cfg.Nx, run_cfg.Ny, run_cfg.N_layers, run_cfg.N_freq],
        dtype=torch.cfloat,
    )

    # Net out is [0,1] thus we transform to desired real and imaginary ranges
    # first half contains real entries
    eps.real = out[:, :, :, 0 : run_cfg.N_freq]

    # scale up to [0 , (max-min_eps)]
    eps.real = eps.real * (run_cfg.real_max_eps - run_cfg.real_min_eps)

    # translate to [min_eps, max_eps] and clip outliers
    eps.real = (eps.real + run_cfg.real_min_eps).clip(
        run_cfg.real_min_eps, run_cfg.real_max_eps
    )

    if run_cfg.avoid_zero_eps:
        _avoid_zero_eps(eps)

    # second half imaginary
    eps.imag = out[:, :, :, run_cfg.N_freq :]
    eps.imag = (eps.imag * (run_cfg.imag_max_eps - run_cfg.imag_min_eps)).clip(
        run_cfg.imag_min_eps, run_cfg.imag_max_eps
    )

    # Adds noise scaled to desired eps range times noise_scale
    if run_cfg.add_noise:
        eps.real += (
            (torch.randn_like(eps.real) - 0.5)
            * 2.0
            * run_cfg.noise_scale
            * (run_cfg.real_max_eps - run_cfg.real_min_eps)
        )
        eps.real = eps.real.clip(run_cfg.real_min_eps, run_cfg.real_max_eps)
        eps.imag += (
            (torch.randn_like(eps.imag) - 0.5)
            * 2.0
            * run_cfg.noise_scale
            * (run_cfg.imag_max_eps - run_cfg.imag_min_eps)
        )
        eps.imag = eps.imag.clip(run_cfg.imag_min_eps, run_cfg.imag_max_eps)

    return eps


def _classification_model_to_eps_grid(model, run_cfg: DotMap):
    """Computes a 4D grid of epsilons for a classification model.

    Args:
        model (torch.model): Trained neural network model. Should map one 4D input to a material index.
        run_cfg (DotMap): Configuration for the run.

    Returns:
        [torch.tensor]: Resulting 4D [real,imag] epsilon grid
    """

    Nx_undersampled = run_cfg.Nx // run_cfg.eps_oversampling
    Ny_undersampled = run_cfg.Ny // run_cfg.eps_oversampling

    # Evaluate the model at the grid points
    out = _eval_model(
        model,
        Nx_undersampled,
        Ny_undersampled,
        run_cfg.N_layers,
        run_cfg.target_frequencies,
        run_cfg.freq_distribution,
        run_cfg.PER_LAYER_THICKNESS,
    )

    # Reshape the output to have a 4D tensor again
    # Note we output a likelihood for each material
    out = out.reshape(
        Nx_undersampled,
        Ny_undersampled,
        run_cfg.N_layers,
        run_cfg.N_freq,
        run_cfg.N_materials,
    )

    # Oversample the created grid
    out = torch.repeat_interleave(out, run_cfg.eps_oversampling, dim=0)
    out = torch.repeat_interleave(out, run_cfg.eps_oversampling, dim=1)

    # We take mean over the frequency dimension in determining the fitting
    out = torch.mean(out, dim=3)

    # Adds noise according to specified noise_scale
    if run_cfg.add_noise:
        out += torch.randn_like(out) * run_cfg.noise_scale

    beta = 16
    # Softmax with a high beta to push towards 1
    exponential = torch.exp(beta * out)
    material_id = exponential / exponential.sum(dim=-1).unsqueeze(-1)

    material_id = torch.divide(material_id, material_id.sum(-1).unsqueeze(-1))

    # Convert material probabilities to epsilons
    eps = (material_id.unsqueeze(-1) * run_cfg.material_collection.epsilon_matrix).sum(
        -2
    )

    if run_cfg.avoid_zero_eps:
        _avoid_zero_eps(eps)

    return eps, material_id


def model_to_eps_grid(model, run_cfg: DotMap):
    """Computes a 4D grid of epsilons for the given model.

    Args:
        model (torch.model): Trained neural network model. Should map one 4D input to a [real,imag] epsilon.
        run_cfg (DotMap): Configuration for the run.

    Returns:
        torch.tensor, torch.tensor: Resulting 4D [real,imag] epsilon grid and tensor of the material_ids (None for regression)
    """
    if run_cfg.type == "classification":
        return _classification_model_to_eps_grid(model, run_cfg)
    elif run_cfg.type == "regression":
        # For consistent returns we return a tuple with None
        return _regression_model_to_eps_grid(model, run_cfg), None



import torch
import numpy as np
from matplotlib import pyplot as plt

from ..training.model.model_to_eps_grid import model_to_eps_grid


def plot_model_grid(run_cfg, save_path=None):
    """Plots the absolute value of the epsilon over all frequencies for each 3-D grid point. Optionally saves it.
    Args:
        run_cfg (dict): The run configuration.
        save_path (str, optional): Folder to save the plot in. Defaults to None, then the plot will not be saved.
    """
    Nx, Ny, N_layers = run_cfg.Nx, run_cfg.Ny, run_cfg.N_layers
    eps, _ = model_to_eps_grid(run_cfg.model, run_cfg)
    x = torch.linspace(-1, 1, Nx)
    y = torch.linspace(-1, 1, Ny)
    z = torch.linspace(-1, 1, N_layers)
    X, Y, Z = torch.meshgrid((x, y, z))

    # Here we calculate the absolute value of the permittivity over all frequencies for each grid point
    eps = torch.mean(eps, dim=3)
    eps = torch.sqrt(eps.real ** 2 + eps.imag ** 2)

    abs_values = eps.detach().cpu().numpy()

    X = X.cpu().numpy()
    Y = Y.cpu().numpy()
    Z = Z.cpu().numpy()

    # Here we plot it
    fig = plt.figure(figsize=(4, 4), dpi=150)
    fig.patch.set_facecolor("white")

    ax = fig.add_subplot(111, projection="3d")
    ax.view_init(elev=25, azim=100)
    p = ax.scatter(
        X.reshape(-1, 1),
        Y.reshape(-1, 1),
        Z.reshape(-1, 1),
        marker="s",
        s=120,
        linewidths=0,
        alpha=1.0,
        c=abs_values,
    )
    cbar = plt.colorbar(p, ax=ax)
    cbar.ax.tick_params(labelsize=7)
    ax.grid(False)  # Hide grid lines
    ax.set_xlabel("$N_x =$" + str(Nx))
    ax.set_ylabel("$N_y =$" + str(Ny))
    # ax.set_zlabel("# of layers", rotation=60)  # TODO Fix rotation
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_zticks(np.linspace(-1, 1, N_layers))
    z_labels = [""] * N_layers
    for idx in range(N_layers):
        z_labels[idx] = "L" + str(idx + 1)
    ax.set_zticklabels(
        z_labels
    )  # Where L1 is (seemingly) the bottom one ( #TODO confirm this)
    cbar.set_label("Complex norm of the permittivity", labelpad=10)

    if save_path is not None:
        plt.savefig(save_path + "/model_grid.png", dpi=150)



from dotmap import DotMap

import torch
import numpy as np
from collections import deque
from copy import deepcopy
from loguru import logger

from .losses.spectrum_loss import _spectrum_loss_fn
from .losses.likelihood_regularization_loss import _likelihood_regularization_loss_fn
from ..materials.material_collection import MaterialCollection
from .model.init_network import init_network
from .model.model_to_eps_grid import model_to_eps_grid
from ..utils.compute_spectrum import compute_spectrum
from ..trcwa.compute_target_frequencies import compute_target_frequencies
from ..utils.fix_random_seeds import fix_random_seeds
from ..utils.validate_config import _validate_config


def _init_training(run_cfg: DotMap):
    """Initializes additional parameters required for training.
    Args:
        run_cfg (DotMap): Run configuration.

    Returns:
        DotMap, torch.opt, torch.scheduler: Run config with additional entries and model, optimizer, and scheduler
    """

    # Validate config
    _validate_config(run_cfg)

    # Enable GPU if desired
    if run_cfg.use_gpu:
        torch.set_default_tensor_type(torch.cuda.FloatTensor)

    # Determine target frequencies
    run_cfg.target_frequencies = compute_target_frequencies(
        run_cfg.physical_wavelength_range[0],
        run_cfg.physical_wavelength_range[1],
        run_cfg.N_freq,
        run_cfg.freq_distribution,
    )

    logger.debug("Computed target frequencies:")
    logger.debug(run_cfg.target_frequencies)

    if run_cfg.type == "classification":
        run_cfg.material_collection = MaterialCollection(run_cfg.target_frequencies)
        run_cfg.N_materials = run_cfg.material_collection.N_materials
        # If classification, the model outputs a likelihood for the presence of each material.
        run_cfg.out_features = run_cfg.N_materials
    else:
        run_cfg.out_features = 2  # For regression, the model outputs epsilon directly.

    # Fix random seed for reproducibility
    fix_random_seeds(run_cfg.seed)

    # Init model
    if not "model" in run_cfg.keys() or run_cfg.model is None:
        run_cfg.model = init_network(run_cfg)

    # Initialize some utility
    optimizer = torch.optim.Adam(run_cfg.model.parameters(), lr=run_cfg.learning_rate)
    scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
        optimizer, factor=0.66, patience=500, min_lr=1e-6, verbose=True
    )

    return run_cfg, optimizer, scheduler


def run_training(
    run_cfg: DotMap,
):
    """Runs a training run with the passed config.

    Args:
        run_cfg (DotMap): Run configuration.

    Returns:
        DotMap: The loss results of the training run
    """
    logger.trace("Initializing training...")

    # Initialize training parameters, model and optimizer etc.
    run_cfg, optimizer, scheduler = _init_training(run_cfg)

    run_cfg.results = DotMap()
    # run_cfg.model = model

    # When a new network is created we init empty training logs
    run_cfg.results.loss_log = []
    run_cfg.results.weighted_average_log = []
    run_cfg.results.L1_errs = []
    weighted_average = deque([], maxlen=20)

    # And store the best results
    best_loss = np.inf
    run_cfg.best_model_state_dict = run_cfg.model.state_dict()

    logger.trace("Starting training...")
    # Training Loop
    for it in range(run_cfg.iterations):
        torch.cuda.empty_cache()

        # Compute the epsilon values from the model
        eps_grid, material_ids = model_to_eps_grid(run_cfg.model, run_cfg)

        # Compute the spectrum for this material
        try:
            # Not validating because this already happened in _init_training
            produced_R_spectrum, produced_T_spectrum = compute_spectrum(
                eps_grid, run_cfg, validate_cfg=False
            )
        except ValueError:
            logger.warning(
                "ValueError encountered in compute_spectrum. This likely means the LR was too high. Reloading best model, and reducing LR."
            )
            run_cfg.model.load_state_dict(run_cfg.best_model_state_dict)
            logger.info(
                "Setting LR to {}".format(optimizer.param_groups[0]["lr"] * 0.5)
            )
            optimizer.param_groups[0]["lr"] *= 0.5
            continue
        # Compute loss between target spectrum and
        # the one from the current network structure
        spectrum_loss, renormalized = _spectrum_loss_fn(
            produced_R_spectrum,
            produced_T_spectrum,
            run_cfg.target_reflectance_spectrum,
            run_cfg.target_transmittance_spectrum,
            run_cfg.target_frequencies,
            run_cfg.L,
            run_cfg.absorption_loss,
        )

        # Compute L1 error for logging
        L1err, _ = _spectrum_loss_fn(
            produced_R_spectrum,
            produced_T_spectrum,
            run_cfg.target_reflectance_spectrum,
            run_cfg.target_transmittance_spectrum,
            run_cfg.target_frequencies,
            1,
            False,
        )

        loss = 0
        loss += spectrum_loss

        spectrum_loss = spectrum_loss.item()  # Convert to scalar

        if run_cfg.type == "classification" and run_cfg.use_regularization_loss:
            loss += run_cfg.reg_loss_weight * _likelihood_regularization_loss_fn(
                material_ids, run_cfg.L
            )
        # We store the model if it has the lowest loss yet
        # (this is to avoid losing good results during a run that goes wild)
        if loss < best_loss:
            best_loss = loss
            logger.info(
                f"###  New Best={loss.item():<6.4f} with SpectrumLoss={spectrum_loss:<6.4f} ### L1={L1err.detach().item():.4f}"
            )
            if not renormalized:
                logger.debug("Saving model state...")
                run_cfg.best_model_state_dict = deepcopy(run_cfg.model.state_dict())

        # Update the loss trend indicators
        weighted_average.append(loss.item())

        # Update the logs
        run_cfg.results.weighted_average_log.append(np.mean(weighted_average))
        run_cfg.results.loss_log.append(loss.item())
        run_cfg.results.L1_errs.append(L1err.detach().item())

        # Print every i iterations
        if it % 5 == 0:

            wa_out = np.mean(weighted_average)
            logger.info(
                f"It={it:<5} Loss={loss.item():<6.4f}   |  weighted_avg={wa_out:<6.4f}  |  SpectrumLoss={spectrum_loss:<6.4f} | L1={L1err.detach().item():.4f}"
            )

        # Zeroes the gradient (otherwise would accumulate)
        optimizer.zero_grad()

        # Compute gradient of the loss with respect to model parameters
        loss.backward()

        # Update model parameters
        optimizer.step()

        # Perform a step in LR scheduler to update LR if necessary
        scheduler.step(loss.item())

    logger.trace("Reloading best model state...")
    # Return best model in the end
    run_cfg.model.load_state_dict(run_cfg.best_model_state_dict)

    return run_cfg



from loguru import logger

from ..trcwa.compute_target_frequencies import compute_target_frequencies
from ..utils.global_constants import UNIT_MAGNITUDE


def test_conversion():
    """Test whether conversion from real wavelengths into target frequencies
    is correctly done.
    """
    logger.trace("Running frequency conversion test...")
    min_wl = 300e-9  # 300 nm
    max_wl = 3e-6  # 3 µm

    target_frequencies = compute_target_frequencies(min_wl, max_wl, 5, "linear")
    logger.debug("Proposed frequencies are")
    logger.debug(target_frequencies)

    assert len(target_frequencies) == 5

    prod_min_wl = 1.0 / target_frequencies[-1] * UNIT_MAGNITUDE
    prod_max_wl = 1.0 / target_frequencies[0] * UNIT_MAGNITUDE

    assert abs(prod_min_wl - min_wl) < 1e-8
    assert abs(prod_max_wl - max_wl) < 1e-8


if __name__ == "__main__":
    test_conversion()



from dotmap import DotMap
from tqdm import tqdm
from loguru import logger
import torch

from nidn.fdtd_integration.constants import FDTD_GRID_SCALE
from nidn.utils.global_constants import UNIT_MAGNITUDE

from ..trcwa.get_frequency_points import get_frequency_points
from .calculate_transmission_reflection_coefficients import (
    calculate_transmission_reflection_coefficients,
)
from .init_fdtd import init_fdtd
from ..fdtd.backend import backend as bd


def compute_spectrum_fdtd(permittivity, cfg: DotMap):
    """Generates a spectrum of transmission and reflection coefficients for the specified wavelengths in the cfg, by using FDTD simulations.

    Args:
        permitivity (torch.tensor): Array of permittivity for each layer
        cfg (DotMap): Configurations needed to perform the simulations

    Returns:
        tuple[array, array]: Transmission spectrum and reflection spectrum
    """
    transmission_spectrum = []
    reflection_spectrum = []
    physical_wavelengths, _ = get_frequency_points(cfg)
    logger.debug("Wavelenghts in spectrum : ")
    logger.debug(physical_wavelengths)
    logger.debug("Number of layers: " + str(len(permittivity[0, 0, :, 0])))

    # For each wavelength, calculate transmission and reflection coefficents
    disable_progress_bar = logger._core.min_level >= 20
    for i in tqdm(range(len(physical_wavelengths)), disable=disable_progress_bar):
        logger.debug("Simulating for wavelenght: {}".format(physical_wavelengths[i]))
        transmission_signal = []
        reflection_signal = []

        # Create simulation in free space and run it
        grid, transmission_detector, reflection_detector = init_fdtd(
            cfg,
            include_object=False,
            wavelength=physical_wavelengths[i],
            permittivity=None,
        )
        grid.run(cfg.FDTD_niter, progress_bar=False)
        transmission_free_space, reflection_free_space = _get_detector_values(
            transmission_detector, reflection_detector
        )
        transmission_signal.append(transmission_free_space)
        reflection_signal.append(reflection_free_space)

        # Create the same simulation, but add material in the form of one or many layers, and run again
        grid, transmission_detector, reflection_detector = init_fdtd(
            cfg,
            include_object=True,
            wavelength=physical_wavelengths[i],
            permittivity=permittivity[:, :, :, i],
        )
        grid.run(cfg.FDTD_niter, progress_bar=False)
        transmission_material, reflection_material = _get_detector_values(
            transmission_detector, reflection_detector
        )
        transmission_signal.append(transmission_material)
        reflection_signal.append(reflection_material)
        # Calculate transmission and reflection coefficients,
        # by using the signals from the free space simulation and the material simulation
        (
            transmission_coefficient,
            reflection_coefficient,
        ) = calculate_transmission_reflection_coefficients(
            transmission_signal, reflection_signal
        )
        transmission_spectrum.append(transmission_coefficient)
        reflection_spectrum.append(reflection_coefficient)
    logger.debug("Trasmission spectrum")
    logger.debug(transmission_spectrum)
    logger.debug("Reflection spectrum")
    logger.debug(reflection_spectrum)

    return reflection_spectrum, transmission_spectrum


def _get_detector_values(transmission_detector, reflection_detector):
    """Extract the signals detected by the transmission detector and the reflection detector

    Args:
        transmission_detector (fdtd.LineDetector): The transmission detector attached to the grid
        reflection_detector (fdtd.LineDetector): The reflection detector attached to the grid
        detector_position_y (int): The position along the linedetector which should be used

    Returns:
        tuple[array, array]: The electric field detected by the transmission detector, and the electric field detected by the reflection detector
    """
    e_transmission = _get_abs_value_from_3D_signal(
        transmission_detector.detector_values()["E"]
    )
    e_reflection = _get_abs_value_from_3D_signal(
        reflection_detector.detector_values()["E"]
    )
    h_transmission = _get_abs_value_from_3D_signal(
        transmission_detector.detector_values()["H"]
    )
    h_reflection = _get_abs_value_from_3D_signal(
        reflection_detector.detector_values()["H"]
    )
    return e_transmission, e_reflection


def _get_abs_value_from_3D_signal(signal):
    """Get the absolute value of a three-dimentional signal. For each time step, a =sqrt(x**2 + y**2 + z**2)

    Args:
        signal (array): Three dimentional time signal

    Returns:
        Array: One dimentional time-signal
    """
    signal = _average_along_detector(signal)

    abs_value = bd.zeros(len(signal))
    for i in range(len(signal)):
        # Added 1e-16 to prevent gradient flow from breaking, without significantly changing the result
        squared_value = torch.square(signal[i] + 1e-16)
        summed_squared_value = torch.sum(squared_value)
        absolute_value = torch.sqrt(summed_squared_value)
        abs_value[i] = absolute_value
    # For now only returning z-component of electric field, because the signal is only in the z-direction.
    # The absoulte value of the three directions might be neccesary later if the electric field is present in other directions.
    return signal[:, 2]


def _average_along_detector(signal):
    """Average the signal along each point of a line detector

    Args:
        signal (Array[timesteps, points_on_detector,3]): E or H -field signal from a line detector

    Returns:
        Array[timesteps, 3]: averaged signal along detector
    """
    avg = bd.zeros([len(signal), 3])
    for i in range(len(signal)):
        s = bd.zeros(3)
        for p in signal[i]:
            s[0] += p[0] / len(signal[i])
            s[1] += p[1] / len(signal[i])
            s[2] += p[2] / len(signal[i])
        avg[i] = s
    return avg



from .global_constants import UNIT_MAGNITUDE, SPEED_OF_LIGHT


def freq_to_wl(freq):
    """Gives the wavelength in units of the lattice constants
    from the unitless frequency and vice versa. In TRCWA the magnitude is set to micrometres (um).

    Args:
        freq (float): The unitless frequency, or the wavelength in units of TRCWA_UNIT_MAGNITUDE.

    Returns:
        float: The wavelength with units of TRCWA_UNIT_MAGNITUDE, or the unitless frequency.
    """
    return 1.0 / freq


def phys_wl_to_wl(phys_wl):
    """Gives the wavelength from the physical wavelength.

    Args:
        phys_wl (float): The physical wavelength in metres.

    Returns:
        float: The wavelength in units of TRCWA_UNIT_MAGNITUDE.
    """
    return phys_wl / UNIT_MAGNITUDE


def wl_to_phys_wl(wl):
    """Gives the physical wavelength from the wavelength in units of TRCWA_UNIT_MAGNITUDE.

    Args:
        wl (float): The wavelength in units of TRCWA_UNIT_MAGNITUDE.

    Returns:
        float: The physical wavelength in metres.
    """
    return wl * UNIT_MAGNITUDE


def phys_freq_to_phys_wl(freq):
    """Gives the physical frequency from the physical wavelength and vice versa.

    Args:
        input (float): The physical wavelength in metres or the physical frequency in Hz.

    Returns:
        float: The physical frequency or the physical wavelength.
    """
    c = SPEED_OF_LIGHT
    return c / freq



import torch
import numpy as np
from matplotlib import pyplot as plt
from ..utils.convert_units import freq_to_wl
from ..training.model.model_to_eps_grid import model_to_eps_grid


def plot_model_grid_per_freq(run_cfg, freq_idx=[0, 1, 2, 3], save_path=None):
    """Plots the real and imaginary part of the permittivity in two separate plots for
    frequencies in target_frequencies defined by freq_idx. Optionally saves it.

    Args:
        run_cfg (dict): The run configuration.
        freq_idx (list of int): Which of the frequency indices in target_frequencies we want to plot. Defaults to [0, 1, 2, 3].
        save_path (str, optional): Folder to save the plot in. Defaults to None, then the plot will not be saved.
    """
    Nx, Ny, N_layers = run_cfg.Nx, run_cfg.Ny, run_cfg.N_layers
    eps, _ = model_to_eps_grid(run_cfg.model, run_cfg)

    x = torch.linspace(-1, 1, Nx)
    y = torch.linspace(-1, 1, Ny)
    z = torch.linspace(-1, 1, N_layers)
    X, Y, Z = torch.meshgrid((x, y, z))

    X = X.cpu().numpy()
    Y = Y.cpu().numpy()
    Z = Z.cpu().numpy()

    # Here we plot it
    fig = plt.figure(figsize=(8, 12), dpi=150)
    fig.patch.set_facecolor("white")

    min_eps_real, max_eps_real = eps.real.min(), eps.real.max()
    min_eps_imag, max_eps_imag = eps.imag.min(), eps.imag.max()

    for idx in range(len(freq_idx)):
        eps_real = eps[:, :, :, freq_idx[idx]].real
        eps_imag = eps[:, :, :, freq_idx[idx]].imag
        material_id_real = eps_real.detach().cpu().numpy()
        material_id_imag = eps_imag.detach().cpu().numpy()

        re_or_im_idx = 0
        for real_or_imag in [material_id_real, material_id_imag]:
            ax = fig.add_subplot(421 + re_or_im_idx + idx * 2, projection="3d")
            ax.view_init(elev=25, azim=100)
            vmin, vmax = min_eps_real, max_eps_real
            if re_or_im_idx > 0:
                vmin, vmax = min_eps_imag, max_eps_imag
            p = ax.scatter(
                X.reshape(-1, 1),
                Y.reshape(-1, 1),
                Z.reshape(-1, 1),
                marker="s",
                linewidths=0,
                s=140,
                alpha=1.0,
                c=real_or_imag,
                vmin=vmin,
                vmax=vmax,
            )
            ax.set_title(
                "WL="
                + str(round(freq_to_wl(run_cfg.target_frequencies[freq_idx[idx]]), 2))
                + "um",
                fontsize=8,
            )
            # ax.set_xlabel("$N_x =$" + str(Nx))
            # ax.set_ylabel("$N_y =$" + str(Ny))
            # ax.set_zlabel("# of layers", rotation=60)  # TODO Fix rotation
            # ax.grid(False)  # Hide grid lines
            ax.set_xticks([])
            ax.set_yticks([])
            ax.set_zticks(np.linspace(-1, 1, N_layers))
            z_labels = [""] * N_layers
            for layer in range(N_layers):
                z_labels[layer] = "L" + str(layer + 1)
            ax.set_zticklabels(
                z_labels
            )  # Where L1 is (seemingly) the bottom one ( #TODO confirm this)
            cb = plt.colorbar(p, ax=ax)
            cb.set_label("Real permittivity, $\epsilon'$", rotation=270, labelpad=15)
            if re_or_im_idx == 1:
                cb.set_label(
                    "Imaginary permittivity, $\epsilon''$", rotation=270, labelpad=15
                )
            cb.ax.tick_params(labelsize=7)
            re_or_im_idx += 1

    if save_path is not None:
        plt.savefig(save_path + "/model_grid_per_freq.png", dpi=150)



from loguru import logger
import sys


def set_log_level(log_level: str):
    """Set the log level for the logger.

    Args:
        log_level (str): The log level to set. Options are 'TRACE','DEBUG', 'INFO', 'SUCCESS', 'WARNING', 'ERROR', 'CRITICAL'.
    """
    logger.remove()
    logger.add(
        sys.stderr,
        colorize=True,
        level=log_level,
        format="<green>{time:HH:mm:ss}</green>|NIDN-<blue>{level}</blue>| <level>{message}</level>",
        filter="nidn",
    )
    logger.debug(f"Setting LogLevel to {log_level}")



import numpy as np
import torch

from ..constants import *
from ...utils.global_constants import PI


def Lattice_Reciprocate(L1, L2):
    """Given two lattice vectors L1,L2 in the form of (Lx,Ly), returns the
    reciprocate vectors Lk/(2*pi)
    """

    assert (
        type(L1) == list and type(L2) == list
    ), "Lattice vectors should be in list format."
    assert len(L1) == 2, "Both x,y components of Lattice vector L1 are required."
    assert len(L2) == 2, "Both x,y components of Lattice vector L2 are required."

    d = L1[0] * L2[1] - L1[1] * L2[0]

    Lk1 = torch.tensor([L2[1] / d, -L2[0] / d])
    Lk2 = torch.tensor([-L1[1] / d, L1[0] / d])

    return Lk1, Lk2


def Lattice_getG(nG, Lk1, Lk2, method=0):
    """
    The G is defined to produce the following reciprocal vector:
    k = G[:,0] Lk1 + G[:,1] Lk2 (both k and Lk don't include the 2pi factor)

    method:0 for circular truncation, 1 for parallelogramic truncation
    """
    assert type(nG) == int, "nG must be integer"

    if method == 0:
        G, nG = Gsel_circular(nG, Lk1, Lk2)
    elif method == 1:
        G, nG = Gsel_parallelogramic(nG, Lk1, Lk2)
    else:
        raise Exception("Truncation scheme is not included")

    return G, nG


def Lattice_SetKs(G, kx0, ky0, Lk1, Lk2):
    """
    Construct kx,ky including all relevant orders, given initial scalar kx,ky
    2pi factor is now included in the returned kx,ky
    """

    kx = kx0 + 2 * PI * (Lk1[0] * G[:, 0] + Lk2[0] * G[:, 1])
    ky = ky0 + 2 * PI * (Lk1[1] * G[:, 0] + Lk2[1] * G[:, 1])

    return kx, ky


def Gsel_parallelogramic(nG, Lk1, Lk2):
    """From Liu's gsel.c"""
    u = torch.norm(Lk1)
    v = torch.norm(Lk2)
    uv = torch.dot(Lk1, Lk2)

    NGroot = int(torch.sqrt(nG))
    if torch.mod(NGroot, 2) == 0:
        NGroot -= 1

    M = NGroot // 2

    xG = range(-M, NGroot - M)
    G1, G2 = torch.meshgrid(xG, xG, indexing="ij")
    G1 = G1.flatten()
    G2 = G2.flatten()

    # sorting
    Gl2 = G1**2 * u**2 + G2**2 * v**2 + 2 * G2 * G1 * uv
    sort = torch.argsort(Gl2)
    G1 = G1[sort]
    G2 = G2[sort]

    # final G
    nG = NGroot * NGroot
    G = torch.zeros((nG, 2), dtype=int)
    G[:, 0] = G1[:nG]
    G[:, 1] = G2[:nG]

    return G, nG


def Gsel_circular(nG, Lk1, Lk2):
    """From Liu's gsel.c.
    NG * abs(u x v) is approximately the area in k-space we will need
    cover with a circular disc. (u and v are the 2 shortest lattice
    vectors) From the area, we can find the radius (and round it
    up). Then, we can find the minimum extends in each of the two
    lattice directions.
    """
    u = torch.norm(Lk1)
    v = torch.norm(Lk2)
    uv = torch.dot(Lk1, Lk2)
    uxv = Lk1[0] * Lk2[1] - Lk1[1] * Lk2[0]
    circ_area = nG * torch.abs(uxv)
    circ_radius = torch.sqrt(circ_area / PI) + u + v

    u_extent = 1 + int(circ_radius / (u * torch.sqrt(1.0 - uv**2 / (u * v) ** 2)))
    v_extent = 1 + int(circ_radius / (v * torch.sqrt(1.0 - uv**2 / (u * v) ** 2)))

    uext21 = 2 * u_extent + 1
    vext21 = 2 * v_extent + 1

    xG = range(-u_extent, uext21 - u_extent)
    yG = range(-v_extent, vext21 - v_extent)
    G1, G2 = torch.meshgrid(torch.tensor(xG), torch.tensor(yG))
    G1 = G1.flatten()
    G2 = G2.flatten()

    # sorting
    Gl2 = G1**2 * u**2 + G2**2 * v**2 + 2 * G2 * G1 * uv

    # Torch's sorting algorithm is not stable, so we sort with numpy for reproducibility
    sort = np.argsort(Gl2.cpu().numpy())
    # sort = torch.argsort(Gl2)

    G1 = G1[sort]
    G2 = G2[sort]
    Gl2 = Gl2[sort]

    nGtmp = uext21 * vext21

    if nG < nGtmp:
        nGtmp = nG

    # removing the part outside the cycle
    tol = 1e-10 * max(u**2, v**2)
    for i in range(nGtmp - 1, -1, -1):
        if torch.abs(Gl2[i] - Gl2[i - 1]) > tol:
            break
    nG = i

    # final G
    G = torch.zeros((nG, 2), dtype=int)
    G[:, 0] = G1[:nG]
    G[:, 1] = G2[:nG]

    return G, nG



import torch
from matplotlib import pyplot as plt
from matplotlib.ticker import AutoLocator, AutoMinorLocator

from ..utils.convert_units import freq_to_wl
from ..training.model.model_to_eps_grid import model_to_eps_grid
from ..trcwa.load_material_data import _load_material_data
from ..materials.material_collection import MaterialCollection
from ..materials.find_closest_material import _find_closest_material
from ..utils.global_constants import NIDN_FONTSIZE, NIDN_PLOT_COLOR_1


def plot_eps_per_point(run_cfg, compare_to_material=None, save_path=None, legend=True):
    """This function plots the epsilon values of grid points against real materials. Optionally saves it.

    Args:
        run_cfg (dict): The run configuration.
        compare_to_material (str or list): Name(s) of the material to compare with. Available ones are in /materials/data.
        save_path (str, optional): Folder to save the plot in. Defaults to None, then the plot will not be saved.
    """
    # Create epsilon grid from the model
    eps, _ = model_to_eps_grid(run_cfg.model, run_cfg)
    eps_np = eps.detach().cpu().numpy()

    material_collection = MaterialCollection(run_cfg.target_frequencies)

    # Load material data for comparison
    if compare_to_material is not None:
        if type(compare_to_material) is str:
            material_data = material_collection[compare_to_material]
        else:
            material_data = []
            for mat in compare_to_material:
                material_data.append(material_collection[mat])

    # Create figure
    fig = plt.figure(figsize=(10, 3), dpi=300)
    fig.patch.set_facecolor("white")

    # Plot epsilon
    ax = fig.add_subplot(121)
    ax2 = fig.add_subplot(122)
    wl = freq_to_wl(run_cfg.target_frequencies)

    # Add some horizontal space
    ax.set_xlim(wl.min(), wl.max() + (0.3 * (wl.max() - wl.min())))
    ax2.set_xlim(wl.min(), wl.max() + (0.3 * (wl.max() - wl.min())))

    ax.set_xlabel("Wavelength [µm]")
    ax.set_ylabel("Permittivity - Real Part")
    # ax.set_xscale("log")

    ax2.set_xlabel("Wavelength [µm]")
    ax2.set_ylabel("Permittivity - Imaginary Part")
    # ax2.set_xscale("log")

    ax.xaxis.set_major_formatter(plt.FuncFormatter(lambda x, pos: "{:.1f}".format(x)))
    ax2.xaxis.set_major_formatter(plt.FuncFormatter(lambda x, pos: "{:.1f}".format(x)))

    ax.xaxis.set_major_locator(AutoLocator())
    ax2.xaxis.set_major_locator(AutoLocator())

    # Iterate over all grid points
    for x in range(eps.shape[0]):
        for y in range(eps.shape[1]):
            for N_layer in range(eps.shape[2]):
                eps_point_real = eps_np[x, y, N_layer].real
                eps_point_imag = eps_np[x, y, N_layer].imag
                ax.plot(wl, eps_point_real, linewidth=2)
                ax.minorticks_off()
                ax2.plot(wl, eps_point_imag, linewidth=2)
                ax2.minorticks_off()

                # if not legend:
                #     ax.text(
                #         wl.max(),
                #         eps_point_real[0],
                #         f" {N_layer},{x},{y}",
                #         va="center",
                #         fontsize=7,
                #     )
                #     ax2.text(
                #         wl.max(),
                #         eps_point_imag[0],
                #         f" {N_layer},{x},{y}",
                #         va="center",
                #         fontsize=7,
                #     )

    if legend:
        # Add legend
        names = [
            f"Layer {N_layer} - (x,y)=({x},{y})"
            for x in range(eps.shape[0])
            for y in range(eps.shape[1])
            for N_layer in range(eps.shape[2])
        ]
        ax.legend(
            names,
            fontsize=NIDN_FONTSIZE - 6,
            bbox_to_anchor=(1.05, 1.2),
            ncol=10,
            loc="upper center",
        )
        # ax2.legend(names, fontsize=NIDN_FONTSIZE - 6)

    # Plot material data
    if compare_to_material is not None:
        if type(compare_to_material) is str:
            material_data = [material_data]
            compare_to_material = [compare_to_material]
        for mat_data, mat_name in zip(material_data, compare_to_material):

            ax.plot(wl, mat_data.real, "--", color="black", linewidth=1.5)
            ax.minorticks_off()
            ax2.plot(wl, mat_data.imag, "--", color="black", linewidth=1.5)
            ax2.minorticks_off()
            ax.text(
                wl.max(),
                mat_data.real[0],
                " " + mat_name,
                va="center",
                fontsize=7,
            )
            ax2.text(
                wl.max(),
                mat_data.imag[0],
                " " + mat_name,
                va="center",
                fontsize=7,
            )
    else:
        _, indices = _find_closest_material(eps, run_cfg, material_collection)
        unique_indices = torch.unique(indices)
        names = [material_collection.material_names[i] for i in unique_indices]
        for name in names:
            material_data = material_collection[name]
            ax.plot(wl, material_data.real, "--", color="black", linewidth=1.5)
            ax2.plot(wl, material_data.imag, "--", color="black", linewidth=1.5)
            ax.text(
                wl.max(), material_data.real[0], " " + name, va="center", fontsize=7
            )
            ax2.text(
                wl.max(), material_data.imag[0], " " + name, va="center", fontsize=7
            )

    fig.autofmt_xdate()

    if save_path is not None:
        plt.savefig(save_path + "/eps_per_points.png", dpi=150)



from loguru import logger

from .constants import *
from ..utils.global_constants import *
from .trcwa import TRCWA
from ..materials.material_collection import MaterialCollection


def _init_trcwa(eps_grid, target_frequency, run_cfg):
    """Creates a TRCWA object matching the given eps_grid and target_frequency.

    Args:
        eps_grid (torch.tensor): Grid of epsilon values. Should be [Nx, Ny, N_layers].
        target_frequency (float): Target frequency for this simulation run.
        run_cfg (DotMap): Run configuration.

    Returns:
        TRCWA obj: The created object which is ready to compute the spectrum
    """
    Nx, Ny, N_layers = eps_grid.shape[0:3]

    # Squeeze out Nx=1, Ny=1 dimension (for uniform layer)
    eps_grid = eps_grid.squeeze(0)  # squeeze X
    eps_grid = eps_grid.squeeze(0)  # squeeze Y (now at 0)

    # Adding a small imaginary part to the frequency to avoid singularities in RCWA
    # See page 2 of arxiv.org/pdf/2005.04840v1.pdf
    # or doi.org/10.1364/OE.21.030812 for a more thorough explanation
    freqcmp = target_frequency * (1 + (1j / (2.0 * TRCWA_Q_ABS)))

    if torch.tensor(run_cfg.TRCWA_L_grid).max() * target_frequency > 3.0:
        logger.warning(
            f"With a frequency of {target_frequency} and L_grid={run_cfg.TRCWA_L_grid} TRCWA may become unstable. Consider decreasing the target frequency or decreasing L_grid if R+T+A > 1."
        )

    # Initialize TRCWA object
    trcwa = TRCWA(
        run_cfg.TRCWA_NG,
        run_cfg.TRCWA_L_grid[0],
        run_cfg.TRCWA_L_grid[1],
        freqcmp,
        TRCWA_THETA,
        TRCWA_PHI,
        verbose=0,
    )

    # Add vacuum layer at the top
    trcwa.Add_LayerUniform(
        thickness=1.0, epsilon=torch.tensor(run_cfg.TRCWA_TOP_LAYER_EPS)
    )

    # Add material layers (homogeneous if Nx and Ny are 1)
    for layer in range(N_layers):

        # Set thickness based on config
        if len(run_cfg.PER_LAYER_THICKNESS) == 1:
            thickness = run_cfg.PER_LAYER_THICKNESS[0]
        else:
            thickness = run_cfg.PER_LAYER_THICKNESS[layer]

        if Nx > 1 or Ny > 1:
            trcwa.Add_LayerGrid(thickness, Nx, Ny)
        else:
            trcwa.Add_LayerUniform(thickness, eps_grid[layer])

    # Add layer at the bottom
    trcwa.Add_LayerUniform(
        thickness=1.0, epsilon=torch.tensor(run_cfg.TRCWA_BOTTOM_LAYER_EPS)
    )

    # Initialize the object properly
    trcwa.Init_Setup()

    trcwa.MakeExcitationPlanewave(
        TRCWA_PLANEWAVE["p_amp"],
        TRCWA_PLANEWAVE["p_phase"],
        TRCWA_PLANEWAVE["s_amp"],
        TRCWA_PLANEWAVE["s_phase"],
        order=0,
    )

    # Apply eps if heterogeneous layers
    if Nx > 1 or Ny > 1:
        trcwa.GridLayer_geteps(eps_grid)

    return trcwa



import pickle as pk
from datetime import datetime
from loguru import logger


def load_run(filename):
    """Loads the saved DotMap run config.

    Args:
        filename (str): Path to file.
    Returns:
        DotMap: Run config
    """
    with open(filename, "rb") as f:
        run = pk.load(f)
    logger.info(f"Loaded run config from {filename}")
    return run



"""Utility functions to make torch operations compatible with numpy syntax."""
import torch


def torch_zeros(shape, dtype):
    """Implementation of torch.zeros that returns a complex tensor. Else falls back on torch.zeros.

    Args:
        shape (list): Shape vector for zeros tensor.
        dtype (dtype): Requested datatype.

    Returns:
        torch.tensor: Initialized tensor with real and imaginary part set to zero
    """
    if dtype == complex:
        real = torch.zeros(shape)
        imag = torch.zeros(shape)
        return torch.complex(real, imag)
    else:
        return torch.zeros(shape, dtype=dtype)


def torch_eye(size, dtype=None):
    """Implementation of torch.eye that returns a complex tensor. Else falls back on torch.eye.

    Args:
        shape (list): Shape vector for identity matrix.
        dtype (dtype): Requested datatype.

    Returns:
        torch.tensor: Initialized tensor with real part of the diagonal set to one and imaginary part of the diagonal set to zero, and zeros elsewhere
    """
    if dtype == complex:
        real = torch.eye(size)
        imag = torch.zeros([size, size])
        return torch.complex(real, imag)
    else:
        return torch.eye(size, dtype=dtype)


def torch_transpose(n_dimensional_tensor):
    """Custom transpose for torch as torch.transpose diverges from numpy, see https://github.com/pytorch/pytorch/issues/51280

    Args:
        n_dimensional_tensor (torch.tensor): Tensor to transpose.

    Returns:
        torch.tensor: Transposed tensor
    """
    # Construct reverse index (e.g. [3,2,1,0])
    indices = [
        len(n_dimensional_tensor.shape) - i - 1
        for i in range(len(n_dimensional_tensor.shape))
    ]
    return n_dimensional_tensor.permute(indices)


def torch_dot(a, b):
    """Computes a dot product in torch that is compatible with numpy style syntax.

    Args:
        a (torch.tensor): Tensor a in the dot product of a and b.
        b (torch.tensor): Tensor b in the dot product of a and b.

    Returns:
        torch.tensor: Dot product of a and b
    """
    if a.dtype != b.dtype:  # in case data types are different
        if a.dtype == torch.cfloat or a.dtype == torch.cdouble:
            b = b.type(a.dtype)
        elif b.dtype == torch.cfloat or b.dtype == torch.cdouble:
            a = a.type(b.dtype)
    # to be compatible with numpy broadcasting
    if len(a.shape) == 2 and len(b.shape) == 1:
        return torch.mv(a, b)
    if len(a.shape) > 1:
        return torch.mm(a, b)
    else:
        return torch.dot(a, b)



import os
import torch
from loguru import logger

# Add exposed features here
from .materials.layer_builder import LayerBuilder
from .materials.material_collection import MaterialCollection
from .plots.plot_epsilon_grid import plot_epsilon_grid
from .plots.plot_eps_per_point import plot_eps_per_point
from .plots.plot_losses import plot_losses
from .plots.plot_material_grid import plot_material_grid
from .plots.plot_model_grid import plot_model_grid
from .plots.plot_model_grid_per_freq import plot_model_grid_per_freq
from .plots.plot_spectra import plot_spectra
from .plots.plot_spectrum import plot_spectrum
from .plots.save_all_plots import save_all_plots
from .training.load_run import load_run
from .training.model.model_to_eps_grid import model_to_eps_grid
from .training.run_training import run_training
from .training.save_run import save_run
from .trcwa.compute_target_frequencies import compute_target_frequencies
from .trcwa.get_frequency_points import get_frequency_points
from .utils.compute_spectrum import compute_spectrum
from .utils.convert_units import freq_to_wl, wl_to_phys_wl, phys_freq_to_phys_wl
from .utils.fix_random_seeds import fix_random_seeds
from .utils.load_default_cfg import load_default_cfg
from .utils.print_cfg import print_cfg
from .utils.set_log_level import set_log_level

set_log_level("INFO")

# Set main device by default to cpu if no other choice was made before
if "TORCH_DEVICE" not in os.environ:
    os.environ["TORCH_DEVICE"] = "cpu"

logger.info(f"Initialized NIDN for {os.environ['TORCH_DEVICE']}")
# Set precision (and potentially GPU)
torch.set_default_tensor_type(torch.DoubleTensor)
logger.info("Using double precision")
logger.info(f"Switching log level to warning.")
set_log_level("WARNING")

__all__ = [
    "compute_spectrum",
    "compute_target_frequencies",
    "get_frequency_points",
    "fix_random_seeds",
    "freq_to_wl",
    "LayerBuilder",
    "load_default_cfg",
    "load_run",
    "MaterialCollection",
    "model_to_eps_grid",
    "phys_freq_to_phys_wl",
    "plot_epsilon_grid",
    "plot_eps_per_point",
    "plot_losses",
    "plot_material_grid",
    "plot_model_grid",
    "plot_model_grid_per_freq",
    "plot_spectra",
    "plot_spectrum",
    "print_cfg",
    "run_training",
    "save_all_plots",
    "save_run",
    "set_log_level",
    "wl_to_phys_wl",
]
